/*
 * p7base.c -- ParentheTcl (p7tcl)
 *
 # Copyright (c) 2004-2006 Henry Strickland
 #
 # Permission is hereby granted, free of charge, to any person obtaining a
 # copy of this software and associated documentation files (the "Software"),
 # to deal in the Software without restriction, including without limitation
 # the rights to use, copy, modify, merge, publish, distribute, sublicense,
 # and/or sell copies of the Software, and to permit persons to whom the
 # Software is furnished to do so, subject to the following conditions:
 #
 # The above copyright notice and this permission notice shall be included
 # in all copies or substantial portions of the Software.
 #
 # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 # THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 # OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 # ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 # OTHER DEALINGS IN THE SOFTWARE.
 #
 #         (* http://www.opensource.org/licenses/mit-license.php *)
 *
 */

#include "parenthetcl.h"

#define	ERROR_STRING_MAX 130 /* longer than width of stricks virtual terminal */

void p7tcl_VEC__Cleanup( p7obj obj )
{
    int i;
    p7tcl_VEC_ p= (p7tcl_VEC_)obj;
    if (!p) return;
    assert( p->cls == & p7tcl_VEC__cls );
    for (i= 0; i<p->length; i++) {
	if (p->data[i]) p7tcl_Unlink(p->data[i]);
	p->data[i]= NULL;
    }
}

void p7obj_VEC__Cleanup( p7obj obj )
{
    int i;
    p7obj_VEC_ p= (p7obj_VEC_)obj;
    if (!p) return;
    assert( p->cls == & p7obj_VEC__cls );
    for (i= 0; i<p->length; i++) {
	if (p->data[i]) p7tup_Unlink(p->data[i]);
	p->data[i]= NULL;
    }
}

#define ETERNAL_REFCOUNT 0x40000000 /* looks like LOTS and LOTS of links, so it never goes away */

struct struct_p7cls p7uni_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "uni*", -1, 0, NULL 
};
struct struct_p7cls p7utf_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "utf*", -1, 0, NULL
};
struct struct_p7cls p7byt_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "byt*", -1, 0, NULL
};
struct struct_p7cls p7int_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "int*", -1, 0, NULL
};
struct struct_p7cls p7tcl_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "tcl*", -1, 0, & p7tcl_VEC__Cleanup
};
struct struct_p7cls p7obj_VEC__cls = { 
#ifdef P7_MAGIC
    P7_MAGIC,
#endif
    ETERNAL_REFCOUNT, NULL, "*", -1, 0, & p7obj_VEC__Cleanup
};


#ifdef TCL_MEM_DEBUG

#define P7_MALLOC ckalloc
#define P7_FREE ckfree

#else /*TCL_MEM_DEBUG*/

#define P7_MALLOC Tcl_Alloc
#define P7_FREE Tcl_Free

#endif /*TCL_MEM_DEBUG*/


void* p7malloc_func( char* what, char* file, int line, size_t size )
{
    void* p= P7_MALLOC(size);
#if DEBUG>5
    fprintf(stderr, "m %d %s %s %d -> %lu\n", size, what, file, line, (unsigned long)p );
#endif
    memset( p, 0, size );
    return p;
}

void p7free_func( char* what, char* file, int line, void*p )
{
#if DEBUG>5
    fprintf(stderr, "f %lu %s %s %d\n", (unsigned long)p, what, file, line );
#endif
    P7_FREE(p);
}

void p7exit_proc(ClientData clientData)
{
    p7world world= (p7world) clientData;
}


void p7arrange_exit_handler( p7world world )
{
    Tcl_CreateExitHandler( p7exit_proc, world );
}

/**************  error message constructors *******/

p7err p7set_error_static( p7world world, char* what ) 
{
CHECK
	Tcl_SetResult( world->interp, what, TCL_STATIC );
	return TCL_ERROR;
}
p7err p7set_error_d( p7world world, char* what, int x ) 
{
CHECK
    char* buf= p7malloc( "TCL_DYNAMIC", strlen(what) + 30 );
    assert(buf);
    sprintf( buf, what, x );
	Tcl_SetResult( world->interp, buf, TCL_DYNAMIC );
	return TCL_ERROR;
}
p7err p7set_error_s( p7world world, char* what, char* s ) 
{
CHECK
    char* buf= p7malloc( "TCL_DYNAMIC", strlen(what) + strlen(s) + 10 );
    assert(buf);
    sprintf( buf, what, s );
	Tcl_SetResult( world->interp, buf, TCL_DYNAMIC );
	return TCL_ERROR;
}

/****** UTF string compares ********/

p7int p7cmp( p7tcl a, p7tcl b ) {
    char* ap= a? Tcl_GetString(a): "";
    char* bp= b? Tcl_GetString(b): "";

    int alen= Tcl_NumUtfChars( ap, -1 );
    int blen= Tcl_NumUtfChars( bp, -1 );

    /* Tcl_UtfNcmp only compares specified number of bytes, so use the minimum of both lengths */
    int min= alen<blen? alen : blen;
    int cmp= Tcl_UtfNcmp( ap, bp, min );
    /* fprintf(stderr, "Cmp <%s> <%s> %d\n", ap, bp,  cmp ? cmp : (alen<blen) ? -1 :(alen>blen) ? 1 : 0); */

    /* if initial min bytes same, the shorter string is the lesser */
    return cmp ? cmp : (alen<blen) ? -1 :(alen>blen) ? 1 : 0;
}

p7int p7eq ( p7tcl a, p7tcl b ) { return 0 == p7cmp( a, b ); }
p7int p7ne ( p7tcl a, p7tcl b ) { return 0 != p7cmp( a, b ); }
p7int p7lt ( p7tcl a, p7tcl b ) { return 0 > p7cmp( a, b ); }
p7int p7le ( p7tcl a, p7tcl b ) { return 0 >= p7cmp( a, b ); }
p7int p7gt ( p7tcl a, p7tcl b ) { return 0 < p7cmp( a, b ); }
p7int p7ge ( p7tcl a, p7tcl b ) { return 0 <= p7cmp( a, b ); }

extern Tcl_ObjType tclListType;


/* p7tclcond -- p7class reference is true, empty (i.e. null) is false, otherwise Tcl_GetBooleanFromObj, else False.*/
p7err p7tclcond(p7world world, int* rv, p7tcl tcl)
{
    int err= TCL_OK;
    int z=0;
    *rv= 0; /* tentative */

    if (!tcl) return TCL_OK; /* (p7tcl)NULL is always the same as empty string */

    if (tcl->typePtr== &p7classObjType) {
	*rv= (tcl->internalRep.otherValuePtr != NULL);
	return TCL_OK;
    }

    if (tcl->bytes && tcl->bytes[0]=='\0') return TCL_OK; /* already empty */

    if (tcl->typePtr== &tclListType) {
	/* Optimize by not destroying an internal list representation -- ask it for Length; test for zero */
	err= Tcl_ListObjLength(world->interp, tcl, &z);
	if (err) return err; /* really should not happen since we checked the type! */
	*rv= (z!=0);
	return TCL_OK;
    }

    /* first with NULL interp, avoiding expense of creating error message */
    err= Tcl_GetBooleanFromObj( NULL, tcl, &z );
    if (!err) {
	*rv= (z != 0);
	return TCL_OK;
    }

    /* next we give one more shot at producing the empty string, by asking for GetString */
    if ( Tcl_GetString(tcl)[0] == '\0' ) {
	return TCL_OK;
    }
  
    /* finally we re-run Tcl_GetBooleanFromObj, knowing it should give error message, which we will return. */
    err= Tcl_GetBooleanFromObj( world->interp, tcl, &z );
    if (!err) {
	*rv= (z != 0);
	return TCL_OK;
    }
    return err;
}

/**********  p7int  *********/

p7err p7tcl_to_p7int( p7world world, p7int* rv, Tcl_Obj* x )
{
CHECK
	return Tcl_GetIntFromObj(world->interp, x, rv );
}

p7err p7tcl_from_p7int( p7world world, p7tcl *rv, p7int x)
{
CHECK
	Tcl_Obj* z= Tcl_NewIntObj(x);
	if (!z) return TCL_ERROR;
	Tcl_IncrRefCount(z);

CHECK
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
	return TCL_OK;
}

p7err p7tcl_to_p7uni( p7world world, p7uni* rv, Tcl_Obj* x )
{
CHECK
	int i;
	p7err err= Tcl_GetIntFromObj(world->interp, x, &i );
	*rv= (p7uni)i;
	return err;
}

p7err p7tcl_from_p7uni( p7world world, p7tcl *rv, p7uni x)
{
CHECK
	Tcl_Obj* z= Tcl_NewIntObj(x);
	if (!z) return TCL_ERROR;
	Tcl_IncrRefCount(z);

CHECK
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
	return TCL_OK;
}

p7err p7tcl_to_p7utf( p7world world, p7utf* rv, Tcl_Obj* x )
{
CHECK
	int i;
	p7err err= Tcl_GetIntFromObj(world->interp, x, &i );
	*rv= (p7utf)i;
	return err;
}

p7err p7tcl_from_p7utf( p7world world, p7tcl *rv, p7utf x)
{
CHECK
	Tcl_Obj* z= Tcl_NewIntObj(x);
	if (!z) return TCL_ERROR;
	Tcl_IncrRefCount(z);

CHECK
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
	return TCL_OK;
}

p7err p7tcl_to_p7byt( p7world world, p7byt* rv, Tcl_Obj* x )
{
CHECK
	int i;
	p7err err= Tcl_GetIntFromObj(world->interp, x, &i );
	*rv= (p7byt)i;
	return err;
}

p7err p7tcl_from_p7byt( p7world world, p7tcl *rv, p7byt x)
{
CHECK
	Tcl_Obj* z= Tcl_NewIntObj(x);
	if (!z) return TCL_ERROR;
	Tcl_IncrRefCount(z);

CHECK
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
	return TCL_OK;
}

/**********  p7int*  *********/

p7err p7tcl_to_p7int_VEC_( p7world world, p7int_VEC_* rv, Tcl_Obj* x )
{
    int i;
    int len= 0;
    p7tcl* ptr= NULL;
    p7err err= Tcl_ListObjGetElements( world->interp, x, &len, &ptr );
    if (err) return err;

CHECK
	p7err e= p7int_VEC__New( world, rv, len );  /* frees old *rv */
CHECK
	if (e) return e;
	if (*rv) (*rv)->ref_count= 1;

	for (i=0; i<len; i++) {
		int x;
		e= Tcl_GetIntFromObj(world->interp, ptr[i], &x);
		if (e) {
		    // make sure it gets freed.
		    p7int_VEC__Unlink( (*rv) );
		    *rv = NULL;
		    return e;
		}
		(*rv)->data[i]= x;
	}
CHECK
	return TCL_OK;
}
p7err p7tcl_from_p7int_VEC_( p7world world, p7tcl *rv, p7int_VEC_ x)
{
CHECK
    int i;
    int len= x? x->length: 0;
    p7tcl z= Tcl_NewObj();
    Tcl_IncrRefCount(z);
    p7tcl* objv= (p7tcl*) p7malloc( "TEMP", len * sizeof(p7tcl) );
    assert(objv);
    for (i=0; i<len; i++) {

	objv[i]= Tcl_NewIntObj( x->data[i] );
    }
    Tcl_SetListObj( z, len, objv );
    p7free( "TEMP", (void*) objv);
	if (*rv) Tcl_DecrRefCount(*rv);
    *rv = z;
CHECK
    return TCL_OK;
}

/**********  p7tcl*  *********/

p7err p7tcl_to_p7tcl_VEC_( p7world world, p7tcl_VEC_* rv, Tcl_Obj* x )
{
    int i;
    int len= 0;
    p7tcl* ptr= NULL;
    p7err err= Tcl_ListObjGetElements( world->interp, x, &len, &ptr );
    if (err) return err;

CHECK
	p7err e= p7tcl_VEC__New( world, rv, len );  /* frees old *rv */
CHECK
	if (e) return e;
	if (*rv) (*rv)->ref_count= 1;

	for (i=0; i<len; i++) {
		(*rv)->data[i]= ptr[i];
		if (ptr[i]) p7tcl_Link( ptr[i] );
	}
CHECK
	return TCL_OK;
}
p7err p7tcl_from_p7tcl_VEC_( p7world world, p7tcl *rv, p7tcl_VEC_ x)
{
CHECK
    p7tcl z= NULL;
    int i;
    int len= x? x->length: 0;
    p7tcl* objv= (p7tcl*) p7malloc( "TEMP", len * sizeof(p7tcl) );
    assert(objv);
    for (i=0; i<len; i++) {

	objv[i]= x->data[i];
	if (objv[i]) {
	    // TODO -- why iss onnee kind linked,, and the other nnot???????
	    // //  OKAY -- i ccomment out this link; Tcl_NewListObj() should link them, right?
	    //-------- p7tcl_Link( objv[i] );
	} else {
	    // GOOD -- we dont link, cauz NLO wwill.
	    objv[i]= Tcl_NewObj();
	}
    }
    z= Tcl_NewListObj( len, objv );
    Tcl_IncrRefCount(z);
#ifdef DEBUG
    memset( objv, 0, len * sizeof (p7tcl) ); // NULL the memory for better debug
#endif
    p7free( "TEMP", (void*) objv);
    if (*rv) p7tcl_Unlink(*rv);
    *rv = z;
CHECK
    return TCL_OK;
}

/**********  uni*  *********/

p7err p7tcl_to_p7uni_VEC_( p7world world, p7uni_VEC_* rv, Tcl_Obj* x )
{
	int len=0, i=0;
	Tcl_UniChar* ucp= x? Tcl_GetUnicodeFromObj(x, &len) : NULL;

CHECK
	p7err e= p7uni_VEC__New( world, rv, len );  /* frees old *rv */
CHECK
	if (e) return e;
	if (*rv) (*rv)->ref_count= 1;

	for (i=0; i<len; i++) {
		(*rv)->data[i]= ucp[i];
	}
CHECK
	return TCL_OK;
}
p7err p7tcl_from_p7uni_VEC_( p7world world, p7tcl *rv, p7uni_VEC_ x)
{
CHECK
	p7tcl z= x? Tcl_NewUnicodeObj( x->data, x->length ) : Tcl_NewObj();
    Tcl_IncrRefCount(z);
    if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
CHECK
	return TCL_OK;
}

/**********  utf*  *********/

p7err p7tcl_to_p7utf_VEC_( p7world world, p7utf_VEC_* rv, Tcl_Obj* x )
{
	int len=0;
	char* cp= x? Tcl_GetStringFromObj(x, &len): NULL;

CHECK
	p7err e= p7utf_VEC__New( world, rv, len );  /* frees old *rv */
CHECK
	if (e) return e;
	if (*rv) (*rv)->ref_count= 1;
	assert( cp[len] == '\0' ); /* assert NUL termination, though we don't need it */

	memcpy( (*rv)->data, cp, len );
CHECK
	return TCL_OK;
}
p7err p7tcl_from_p7utf_VEC_( p7world world, p7tcl *rv, p7utf_VEC_ x)
{
CHECK
	p7tcl z= x? Tcl_NewStringObj( (char*) x->data, x->length ) : Tcl_NewObj();
    Tcl_IncrRefCount(z);
    if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
CHECK
	return TCL_OK;
}

/**********  byt*  *********/

p7err p7tcl_to_p7byt_VEC_( p7world world, p7byt_VEC_* rv, Tcl_Obj* x )
{
	int len=0;
	unsigned char* cp= x? Tcl_GetByteArrayFromObj(x, &len): NULL;

CHECK
	p7err e= p7byt_VEC__New( world, rv, len );  /* frees old *rv */
CHECK
	if (e) return e;
	if (*rv) (*rv)->ref_count= 1;

	memcpy( (*rv)->data, cp, len );
CHECK
	return TCL_OK;
}
p7err p7tcl_from_p7byt_VEC_( p7world world, p7tcl *rv, p7byt_VEC_ x)
{
CHECK
	p7tcl z= x? Tcl_NewByteArrayObj( (void*) x->data, x->length ) : Tcl_NewObj();
    Tcl_IncrRefCount(z);
    if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
CHECK
	return TCL_OK;
}

/************ tcl ***********/

p7err p7tcl_to_p7tcl( p7world world, p7tcl* rv, Tcl_Obj* x )
{
CHECK
	if (x) Tcl_IncrRefCount(x);
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = x;
CHECK
	return TCL_OK;
}

p7err p7tcl_from_p7tcl( p7world world, p7tcl *rv, p7tcl x)
{
CHECK
	if (x) Tcl_IncrRefCount(x);
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = x;
CHECK
	return TCL_OK;
}


p7err p7tcl_Assign( p7world world ,p7tcl *rv, p7tcl x ) 
{
CHECK
    if (x) Tcl_IncrRefCount(x);
    if (*rv) Tcl_DecrRefCount (*rv);
    *rv= x;
CHECK
    return TCL_OK;
}

p7err p7tcl_GetAt( p7world world ,p7tcl *rv, p7tcl a, int i ) 
{
CHECK
    p7tcl old= *rv;
    p7err err= Tcl_ListObjIndex( world->interp, a, i, rv );
    if (err) { /* make sure */ *rv= old; return err; }
    if (*rv) Tcl_IncrRefCount(*rv);
    if (old) Tcl_DecrRefCount (old);
CHECK
    return TCL_OK;
}

p7err p7tcl_Subscript( p7world world ,p7tcl **rv, p7tcl* aRef, int i ) 
{
CHECK
    int len= 0;
    p7tcl* ptr= NULL;

    if (Tcl_IsShared( *aRef )) {  
	p7tcl_Unlink(  *aRef  ); 
	*aRef = Tcl_DuplicateObj(  *aRef  ); 
	p7tcl_Link(  *aRef  ); 
    }
    p7err err= Tcl_ListObjGetElements( world->interp, *aRef, &len, &ptr );
    if (err) return err;


    if ( i<0 ) 
	    return p7set_error_static( world, "bounds error: (tcl) subscript negative" );
    if ( len<=i ) 
	    return p7set_error_static( world, "bounds error: (tcl) subscript too big" );

    /* unfortunately, we dont know yet if this will be read or written, 
     * so we must invalidate anyway, and Tcl_DuplicateObj it if shared. 
     * Maybe the next version of P7 should let us know if it's read or written.
     * It's a little hard, if we do more c-like grammar hacks to pexpr.
     */
    Tcl_InvalidateStringRep( *aRef ); 
    *rv = ptr+i;
CHECK
    return TCL_OK;
}


p7err p7tup_Assign( p7world world, p7obj* rv, p7obj x) {
#ifdef DEBUG
fdebugf(stderr, "ASSIGN obj @%lx: %lx <- %lx\n", (ulong)rv, (ulong)(*rv), (ulong)x );
#endif
#ifdef P7_MAGIC
	if (x) assert( x->magic == P7_MAGIC );
	if (*rv) assert( (*rv)->magic == P7_MAGIC );
#endif
CHECK
	if (x) p7tup_Link(x);
CHECK
	if (*rv) p7tup_Unlink(*rv); 
	(*rv)= x;
CHECK
	return TCL_OK;
}

p7err p7tup_New( p7world world, p7obj* rv, p7cls cls ) {
CHECK
	if (*rv) p7tup_Unlink(*rv); 
	(*rv)= (p7obj)p7malloc( cls->clsname, cls->clssize );
CHECK
	if (!(*rv)) return p7set_error_static( world, "Out Of Memory" );
	memset( *rv, 0, cls->clssize );

#ifdef P7_MAGIC
	(*rv)->magic= P7_MAGIC;
#endif
	(*rv)->ref_count= 0;
	(*rv)->cls= cls;
CHECK
	return TCL_OK;
}


#if 0
// killed cauz had Link problem
p7tcl p7tcl_newifnull ( p7tcl x )
{
    return x ? x : Tcl_NewObj();   // TODO -- what about Link????
}
#endif

p7tcl p7tcl_p7peek ( p7obj x )
{
    p7tcl result;

    if ( x ) {
	/* hold up to two integers and a little text */
	char buf[ ERROR_STRING_MAX + 2 * TCL_INTEGER_SPACE ];

	sprintf( buf, "objrefcount=%d", x->ref_count );

	result= Tcl_NewStringObj( buf, -1 );
	p7tcl_Link( result );


    } else {
	result= Tcl_NewStringObj( "NULL", -1 );
	p7tcl_Link( result );
    }	

    return result;
}

p7tcl p7tcl_tclpeek ( p7tcl x )
{
    p7tcl result;

    if ( x ) {
	/* hold up to two integers and a little text */
	char buf[ ERROR_STRING_MAX + 2 * TCL_INTEGER_SPACE ];

	Tcl_ObjType* ot= x -> typePtr;
	if ( ot ) {
	    result= Tcl_NewStringObj( ot->name, -1 );
	} else {
	    result= Tcl_NewStringObj( "NOTYPE", -1 );
	}
	p7tcl_Link( result );

	if ( ot == &p7classObjType ) {
	    p7obj obj= x->internalRep.otherValuePtr;
	    if ( x ) {
		p7cls cls= obj->cls;
		assert(cls);

		Tcl_AppendToObj( result, " ", -1 );
		Tcl_AppendToObj( result, cls->clsname, -1 );

		sprintf( buf, " addr=0x%lx", (unsigned long)x );
		Tcl_AppendToObj( result, buf, -1 );

	    } else {
		Tcl_AppendToObj( result, " NULL", -1 );
	    }
	} else if ( ot == &tclListType) {
	    int llen= 0;
	    Tcl_ListObjLength(NULL, x, &llen);

	    sprintf( buf, " llength=%d", llen );
	    Tcl_AppendToObj( result, buf, -1 );
	}
	    

	if ( x->bytes ) {
	    sprintf( buf, " rc=%d utflen=%d", x->refCount, x->length );
	} else {
	    sprintf( buf, " rc=%d", x->refCount );
	}
	Tcl_AppendToObj( result, buf, -1 );

    } else {
	result= Tcl_NewStringObj( "NULL", -1 );
	p7tcl_Link( result );
    }	

    return result;
}

p7tcl p7tcl_new_linked() {
    /* alternate implementation would link & return literal EMPTY */
    p7tcl z= Tcl_NewObj();
    p7tcl_Link(z);
    return z;
}

void p7freeIntRep( p7tcl tcl )
{
    assert( tcl->typePtr == &p7classObjType );
    if (tcl->internalRep.otherValuePtr) {
	p7obj p= (p7obj) tcl->internalRep.otherValuePtr;
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
#ifdef DEBUG
	assert( p->ref_count > 0);
#endif
	p7tup_Unlink(p);
	tcl->internalRep.otherValuePtr= NULL;
    }
}
void p7dupIntRep( p7tcl src, p7tcl dup )
{   
#ifdef DEBUG
    assert( src->refCount > 0 );
    assert( src->typePtr == &p7classObjType );
#endif
    dup->typePtr= &p7classObjType ;
    dup->internalRep.otherValuePtr= src->internalRep.otherValuePtr;
    if (dup->internalRep.otherValuePtr) {
	p7obj p= (p7obj) dup->internalRep.otherValuePtr;
	p7tup_Link(p);
    }
}
void p7updateString( p7tcl tcl )
{
#ifdef DEBUG
    assert( tcl->typePtr == &p7classObjType );
#endif
    if (tcl->internalRep.otherValuePtr) {

	p7obj p= (p7obj) tcl->internalRep.otherValuePtr;
	char* clsname= p->cls->clsname;

#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
	assert( p->cls->magic == P7_MAGIC );
#endif

	/* Could optimize by storing strlen(clsname) in cls object */
	/* TCL_INTEGER_SPACE (=24) will hold 64-bit hex, plus 4 bytes: @ 0 x \0 */
	tcl->bytes= p7malloc( "STRING_REP", strlen(clsname) + TCL_INTEGER_SPACE + 4 ); 
	sprintf( tcl->bytes, "%s@0x%lx", clsname, (unsigned long)p );
	tcl->length= strlen( tcl->bytes );

    } else {
	/* Is there a more efficient way?  Could we change it to reference a constant NULL p7class Tcl_Obj? */
	tcl->bytes= p7malloc( "STRING_REP", 1 );
	tcl->bytes[0]= '\0';	/* empty string */
	tcl->length= 0;		/* empty string */
    }
}
p7err p7setFromAny( Tcl_Interp* interp, p7tcl tcl )
{
    if ( (! tcl->bytes) || tcl->bytes[0]=='\0' || (tcl->bytes[0]=='0' && tcl->bytes[1]=='\0') ) {
	// empty string OR "0" becomes NULL
	tcl->typePtr= &p7classObjType;
	tcl->internalRep.otherValuePtr= NULL;
	return TCL_OK;
    }
    if (interp) {
	Tcl_SetResult( interp, "ParentheTcl expected a p7class in TCL, but got a non-null non-class string", TCL_STATIC );
    }
    return TCL_ERROR;
}

struct Tcl_ObjType p7classObjType = { 
    "p7class", p7freeIntRep, p7dupIntRep, p7updateString, p7setFromAny
};


p7err p7tcl_to_p7tup( p7world world, p7obj* rv, p7tcl tcl, p7cls cls )
{
    if (*rv) p7tup_Unlink( *rv );
    *rv= NULL;

    if ( ! tcl ) {
	/* empty string produces null, although we usually use "0" */
	/* already done: *rv= NULL; */
    } else if ( tcl->typePtr != & p7classObjType ) {
	p7err e= Tcl_ConvertToType( world->interp, tcl, & p7classObjType );
	if (e) return e;

	/* only NULL can succeed -- must not forge memory addresses from strings! */
	/* already done: *rv= NULL; */

    } else if (tcl->internalRep.otherValuePtr) {
	p7obj p= (p7obj) tcl->internalRep.otherValuePtr;
	if ( p->cls != cls ) {
	    char* buf= p7malloc( "DYNAMIC", ERROR_STRING_MAX + strlen(cls->clsname) + strlen(p->cls->clsname) );
	    sprintf( buf, "ParentheTcl expected type '%s' from Tcl but got type '%s'", 
		    cls->clsname, p->cls->clsname );

	    Tcl_SetResult( world->interp, buf, TCL_DYNAMIC );
	    return TCL_ERROR;
	}
	*rv= p;
	if (*rv) p7tup_Link(*rv);
    } else {
	/* was null internal rep */
	/* already done: *rv= NULL; */
    }
    return TCL_OK;
}
p7err p7tcl_from_p7tup( p7world world, p7tcl *rv, p7obj p)
{
CHECK
	p7tcl z= Tcl_NewObj( );
	if (p) Tcl_InvalidateStringRep(z); /* NULL string rep good for null p */
	z->typePtr= & p7classObjType;
	z->internalRep.otherValuePtr= p;
	if (p) p7tup_Link(p);
	if (*rv) Tcl_DecrRefCount(*rv);
	*rv = z;
	Tcl_IncrRefCount(*rv);
CHECK
	return TCL_OK;
}

void p7world_link(p7world world) {
    world->ref_count ++;
}

p7world Last_p7world;
void p7world_unlink(ClientData cd) {
    int i;
    p7world world= (p7world)cd;

    assert( world );
    assert( world->ref_count > 0 );
    world->ref_count --;

    if ( world->ref_count == 0 ) {
	/* time to die */

	Last_p7world= NULL;

	p7globals_unlink(world);
	p7free( "GLOBALS", (void*) world->globals );

	for (i=0; i < world->num_literals; i++ ) {
	    if ( world->literals[i] ) p7tcl_Unlink( world->literals[i] );
	    world->literals[i] = (p7tcl)NULL;
	}
	p7free( "LITERALS", (void*) world->literals );

	memset( world, 0, sizeof *world );
	p7free( "WORLD", (void*) world );
    }
}


p7err p7initialize( p7world world)
{
    Last_p7world= world;

#ifdef USE_TCL_STUBS
    // Tcl_InitStubs(interp, version, exact);
#endif
    Tcl_RegisterObjType( & p7classObjType );
    return TCL_OK;
}
#if 0
void p7check( char* file, int line )
{
    if ( Last_p7world 
	    && Last_p7world->globals 
	    && Last_p7world->globals->g_errorCode 
	    && Last_p7world->globals->g_errorCode->bytes ) {

	assert( Last_p7world->globals->g_errorCode->bytes < 0x10000000 );
    }
    Tcl_ValidateAllMemory(file,line);
}
#endif



/********************************************************************************************/
/****** Expand these even if not debugging *******/

#undef p7obj_Link
#undef p7obj_Unlink

void p7obj_Link( p7obj p ) {
    if (p) {
#ifdef DEBUG
fdebugf(stderr, "LINK p7tup %lx %d (%s)\n", (ulong)p, p->ref_count, p->cls->clsname );
#endif
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
	assert( p->ref_count >= 0 );
CHECK
	++ p->ref_count;
    }
}

void p7obj_Free( p7obj p ) {
	char* clsname= ( p && p->cls && p->cls->clsname ) ? p->cls->clsname : "?unknownClsName?"; 
#ifdef DEBUG
fdebugf(stderr, "FREE p7tup %lx %d (%s)\n", (ulong)p, p->ref_count, p->cls->clsname );
#endif
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
CHECK
	if (p->cls->cleanup) (p->cls->cleanup)(p);
	memset( p, 0x88, sizeof *p ); /* hi bit set makes negative, will trigger assert( p->ref_count >= 0 ) in _Link & _Unlink */
	p7free( clsname, (void*)p );
CHECK
}

void p7obj_Unlink( p7obj p ) {
    if (p) {
#ifdef DEBUG
fdebugf(stderr, "UNLINK p7tup %lx %d (%s)\n", (ulong)p, p->ref_count, p->cls->clsname );
#endif
#ifdef P7_MAGIC
	assert( p->magic == P7_MAGIC );
#endif
CHECK
	-- p->ref_count;
	assert( p->ref_count >= 0 );
	if (p->ref_count == 0) {
		p7obj_Free(p);
	}
    }
CHECK
}

#undef p7tcl_Link
#undef p7tcl_Unlink

void p7tcl_Link( p7tcl x) 
{
CHECK
    if (x) Tcl_IncrRefCount(x);
CHECK
}

void p7tcl_Unlink( p7tcl x) 
{
CHECK
    if (x) Tcl_DecrRefCount(x);
CHECK
}
