#!/usr/bin/tclsh8.4
#
# p7translate.tcl -- ParentheTcl (p7tcl)
#
# Copyright (c) 2004-2006 Henry Strickland
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
#         (* http://www.opensource.org/licenses/mit-license.php *)

# p7translate.tcl -- Parenthetical Translator -- @yak.net: strick -- 2004-06
# abbrevs
interp alias {} eq {} string equal
interp alias {} ne {} string compare
interp alias {} ll {} llength
interp alias {} sl {} string length
interp alias {} sm {} string match
interp alias {} ie {} info exists
interp alias {} names {} array names
proc 0 list {lindex $list 0}
proc 1 list {lindex $list 1}
proc 2 list {lindex $list 2}
proc 3 list {lindex $list 3}
proc 1..end list {lrange $list 1 end}
proc 2..end list {lrange $list 2 end}
proc not args {
    expr { ! [uplevel 1 $args] }
}

proc llvar {varName} {
    upvar 1 $varName var
    expr {[ie var]? [ll $var]: 0}
}
proc lhas {list item} {expr {[lsearch -exact $list $item]>=0}}
proc incr1 {varName} { ;#  incr the var, or set to 1 if it didnt exist
    upvar 1 $varName v
    if {[ie v]} { incr v } else { set v 0 }
    set v
}
proc lrepeat {item n} {
    set z {}
    for {set i 0} {$i<$n} {incr i} {
	lappend z $item
    }
    set z
}

proc human s {
    # human readable string, not too long
	set s [string range $s 0 100]
	set s [string map {"\n" "\\n" "\t" "\\t"} $s]
	regsub -all {[^ -~]+} $s "?" s
	set s
}

proc ` args {
    # uplevel args, printing before & after
    set lev [info level] 
    set lev [expr { $lev<30 ? $lev : 30 }]
    set c [0 $args]
    puts stderr [human "$c [string repeat < $lev] [1..end $args]"]
    set z [uplevel 1 $args]
    puts stderr [human "$c [string repeat > $lev] $z"]
    set z
}
proc ' args {
    # print args
    set lev [info level] 
    set lev [expr { $lev<30 ? $lev : 30 }]
    puts stderr [human "[string repeat . $lev] $args"]
}
    
proc within {msg body} {
    set e [catch {uplevel 1 $body} what]
    if { $e == 1 } {
		set i $::errorInfo
		set c $::errorCode
		catch { set msg [uplevel 1 [list subst $msg]] }
		append i "\n======> $msg"
		regsub {^[^=]} $what "======> &" what
		regsub {^[^=]} $i "======> &" i
		return -code error -errorinfo $i -errorcode $c $what
    }
    return -code $e $what
}

proc emit-p7check {} {
  puts {
    p7world Last_p7world;

    void p7check( char* file, int line )
    {
#if 0
	if ( Last_p7world
		&& Last_p7world->globals
		&& Last_p7world->globals->g_p7errorCode
		&& Last_p7world->globals->g_p7errorCode->bytes ) {

	    assert( (unsigned long)Last_p7world->globals->g_p7errorCode->bytes < 0x10000000 );
	}
#endif
#if DEBUG>9
	Tcl_ValidateAllMemory(file,line);
#endif
    }
  }
}


proc check {explain args} { ;# check evaluation of $args is true
    if { ![ll $args] || ! [uplevel 1 $args] } {

	# attempt deferred substitutions in explain
	catch { set explain [uplevel 1 [list subst $explain]] } 

	if {![string length $explain]} {
	    set explain "internal error in compiler"
	}
	set msg "======> ERROR: $explain"

        if {[ie ::Llinenum]&&[ie ::Pname]} {append msg "\n======> Error near line $::Llinenum in p7proc $::Pname"}

	if {[info exists ::Ltokin]} {
	    append msg "\n======> Source code following error: [human $::Ltokin]"
	}
	append msg "\n======> (* [human $args] *)"

	return -code error -errorinfo $msg "ERROR: $explain"
    }
}
proc throw {explain} {
    uplevel 1 [list check $explain]
}

	

proc code {cod} {
	append ::Code $cod \n
}
proc push-code {} {
	lappend ::CodeStack $::Code
	set ::Code "/* PUSH \[ */\n"
}
proc pop-code {} {
	append ::Code "/* POP \] */\n"
	set z $::Code
	set ::Code [lindex $::CodeStack end]
	set ::CodeStack [lrange $::CodeStack 0 end-1]
	set z
}
proc push-var {var new} {
    upvar 1 ${var}_Stack stack $var v
    lappend stack $v
    set v $new
}
proc pop-var {var} {
    upvar 1 ${var}_Stack stack $var v
	set z $v
	set v [lindex $stack end]
	set stack [lrange $stack 0 end-1]
	set z
}

############### LEX CODE (from another project, should be renamed & cleaned)
# lex.tcl -- ParentheTcl -- strick@yak.net

proc lex_initok a {
	set ::Llinenum 1
	set ::Ltokin $a
	set ::Ltokout ""
	lex_black
}

proc lex_getchar {} {
    global Lt Ld Ltokin Ltokout Llinenum
	if { ""==$Ltokin } {
		set Lt zz
		set Ld ""
	} elseif {[regexp {^([\\].|.)(.*)$} $Ltokin - Ld Ltokin]} {
	    set Lt ch
	} else {
	    error "problem in lex_getchar"
	}
	append Ltokout $Ld
	regsub -all {[^\n]} $Ld {} newlines
	incr Llinenum [string length $newlines]

}

proc lex_black {} {
    global Lt Ld Ltokin Ltokout Llinenum
	if { ""==$Ltokin } {
		set Lt zz
		set Ld ""
		return
	}

	if {[regexp {^([ \t\v]*)(([][{}\";\n])|(\\.)|(([A-Za-z0-9_]|[:][:])+([<]([<*>,A-Za-z0-9_]|[:][:])+[>])?)|([|][|]|[&][&]|<<|>>>|>>|->|==|!=|<=|>=)|(.)|())(.*)$} \
						$Ltokin 0 ::Lwhite Ld 2 3 4 5 6 - - di 7 Ltokin]
	} {
		switch -regexp -- $Ld {
			^\\\{$	{ set Lt lb }
			^\\\}$	{ set Lt rb }
			^\\\[$	{ set Lt ls }
			^\\\]$	{ set Lt rs }
			^\\\($	{ set Lt lp }
			^\\\)$	{ set Lt rp }
			^\\\"$	{ set Lt dq }
			^\\\;$	{ set Lt sc }
			^\\\\n$	{ set Lt ws }
			^\\n$	{ set Lt nl }
			{^[$]$}	{ set Lt do }
			{^[0-9]+$}	{ set Lt nu }
			{^0x[0-9A-Fa-f]+$}	{ set Lt nu }
			{^([A-Za-z0-9_]|[:][:])+([<]([<*>,A-Za-z0-9_]|[:][:])+[>])?$}	{ set Lt id }
			^$	{ set Lt zz }
			default		{ set Lt ch }
		}
		append Ltokout $Ld
		regsub -all {[^\n]} $Ld {} newlines
		incr Llinenum [string length $newlines]
	} else {
		throw "getok cant grok: $Ltokin"
	}
#' Token-- $Lt '$Ld' '[string range $Ltokin 0 50]' [where]
}

		

############### End of LEX CODE (from another project, should be renamed & cleaned)


array set Prim {int p7int  byt p7byt  utf p7utf  uni p7uni  tcl p7tcl  void void 0 p7zero null p7null }

proc serial {} { incr ::SERIAL }
proc where {} {
    set z {}
    for {set i 0} { $i < [info level] } {incr i} {
	lappend z [lindex [info level -$i] 0]
    }
    set z
} 

proc array-reset args {
	foreach var $args {
		upvar 1 $var a
		catch { unset a }
		set a(x) x
		unset a(x)
	}
}
proc gen-link-var {var} {
    set typ $::Type($var)
    if { $typ eq "tcl" } {
	return "  if ( $var ) { p7tcl_Link( $var ); }\n"
    } elseif {[is-refc-type $typ]} {
	return "  if ( $var ) { p7obj_Link( (p7obj) $var ); }\n"
    } 
}
proc gen-unlink-var {var} {
    set typ $::Type($var)
    if { $typ eq "tcl" } {
	return "  if ( $var ) { p7tcl_Unlink( $var ); $var= (p7tcl)NULL; }\n"
    } elseif {[is-refc-type $typ]} {
	return "  if ( $var ) { p7obj_Unlink( (p7obj) $var ); $var= ([mangle $typ])0; }\n"
    }
}
proc new-cfunc {} {
	array-reset ::Type ::Var

	dcl-type 0 0
	dcl-type null null
	dcl-type void void
	dcl-type never never

	set ::Param {}
	set ::Init "\n"
	set ::Link "\n"
	set ::Code "\n"
	set ::Final "final:\n"
}
proc is-var {name} { 
    # normal var in Var table, or simple global or literal[%d]
    expr { [ info exists ::Var($name) ] || [ regexp {^world->\w+(([[][0-9]+[]])|(->\w+))$} $name ] }
}
proc is-vector-type {type} { regexp {[*]$} $type }
proc is-p7class-type {type} { info exists ::Struct($type)}
proc is-refc-type {type} { expr { $type=="tcl" || [is-p7class-type $type] || [is-vector-type $type] } }
proc is-dyna-type {type} { expr { $type=="tcl" || [is-p7class-type $type] } }
proc new-var {var typ {ini <AUTO>} } {
    if [ie ::Var($var)] { error "local variable '$var' already exists" }
    set ::Type($var) $typ
    set ::Var($var) $typ
    if { $ini eq "<AUTO>" } {
	set mt [mangle $typ]
	append ::Init "  $mt $var= ($mt)0; \n"
    } elseif { $ini eq "<PARAM>" } {
	# dont init params; link them instead
	if {[is-refc-type $typ]} {append ::Link [gen-link-var $var]}
    } else {
	# explicit init string is passed
	append ::Init "  $ini \n"
    }
    if {[is-refc-type $typ]} { 
	append ::Final [gen-unlink-var $var]
    }
    return
}
proc dcl-type {var typ} {
	set ::Type($var) $typ
}
proc identifize x {
    string map { * _VEC_ :: _NS_ < _LT_ > _GT_ , _CO_ } $x
}

proc mangle x {
    # if Memoized
    if {[ie ::Mangle($x)]} {return $::Mangle($x)}

    # due to implementation restrictions,
    #   we make this easier:
    if {[regexp {[*]$} $x]} { ;# if ends in star
	if { $x eq "int*" } { return [set ::Mangle($x) "p7int_VEC_"] }
	if { $x eq "tcl*" } { return [set ::Mangle($x) "p7tcl_VEC_"] }
	if { $x eq "byt*" } { return [set ::Mangle($x) "p7byt_VEC_"] }
	if { $x eq "utf*" } { return [set ::Mangle($x) "p7utf_VEC_"] }
	if { $x eq "uni*" } { return [set ::Mangle($x) "p7uni_VEC_"] }
    }
    set base $x

    # due to implementation restrictions,
    #   we comment a bunch out:
    #regsub {([*])*$} $x {} base
    #regexp {([*])*$} $x tail

    #regsub -all {\m(int|uni|utf|byt|void)\M} $x {p7&} z
    #set z [identifize $z]

    #regsub {(_VEC_)*$} $z {} zbase
    #regexp {(_VEC_)*$} $z ztail
    if {[info exists ::Prim($base)]} {
	return [set ::Mangle($x) "$::Prim($x)"]
	#return [set ::Mangle($x) "$::Prim($base)$ztail"]
    }
    if {[info exists ::Struct($base)]} {
	set z [identifize $x]
	return [set ::Mangle($x) "p7tup_$z"]
    }
    throw "Unknown Type ($x)"
}
proc element-type x {
    if {$x eq "tcl"} {return "tcl"}
    check "element-type of not-a-vector (* $x *)" regsub {[*]$} $x {} x
    set x
}

proc if-err-jump {args} {
    set z "    if (err) \{\n"
    foreach a $args {
	regsub -all {[\'\"\\]} $a {\\&} a
	append z "Tcl_AddErrorInfo( world->interp, \"\\n    $a\" );\n"
    }
    if [info exists ::Where] {
	if [ie ::Llinenum] {
	    append z "Tcl_AddErrorInfo( world->interp, \"\\n$::Where : near line $::Llinenum\" );\n"
	} else {
	    append z "Tcl_AddErrorInfo( world->interp, \"\\n$::Where\" );\n"
	}
    }
    append z "
        goto $::FinalLabel;
    \}"
    set z
}
proc end-of-word {} {
    if { $::Lwhite=="" && $::Lt!="sc" && $::Lt!="nl" && $::Lt!="zz" && $::Lt!="rs" && $::Lt!="rb" } {
    	throw "end-of-word required but saw '$::Ld'"
    }
}
proc p-typespec {} {
	check "typename not a valid identifier (* $::Ld *)" eq $::Lt id
    	set z $::Ld
	lex_black
	while { $::Ld == "*" } {
		append z *
		lex_black
	}
	return $z
}
proc p-opt-paren-typespec {} {
    if { $::Ld == "(" } {
	p-get "("
	set z [p-typespec]
	p-get ")"
	return $z
    } else {
    	return "?"
    }
}
proc p-get-ident {} {
    check "not a valid identifier (* $::Ld : $::Lt *)" eq $::Lt id
    set id $::Ld
    lex_black
    set id
}
proc p-get expected {
    if { $::Ld != $expected } { throw "expected '$expected' got '$::Ld'" }
    lex_black
}

proc parse-typed-id {vartypeName varName spec} {
    upvar 1 $vartypeName vartype $varName var
	if {[regexp {^[(]([A-Za-z0-9_:*<,>]+)[)]([A-Za-z0-9_:*<,>]+)$} $spec - vartype var]} { 
	    return
	} elseif {[regexp {^[A-Za-z0-9_:*<,>]+$} $spec var]} { 
	    set vartype "tcl"
	    return
	} else {
	    check {cannot parse (type)varname syntax: $spec} expr false
	}
}
proc master-p7global {spec} {
    parse-typed-id gt g $spec
    set ::P7Global($g) $gt
    puts stderr "-- p7global  ($gt)$g"
}
proc generate-p7globals {} {
    puts "typedef struct struct_p7globals \{"
    #puts "#ifdef P7_MAGIC"
    #puts "  int magic;"
    #puts "#endif"
    foreach g [array names ::P7Global] {
	set gt $::P7Global($g)
	set mgt [mangle $gt]
	puts "  $mgt [identifize g_$g];"
    }
    puts "\} *p7globals;"

    puts "void p7globals_unlink(p7world world) \{"
    foreach g [array names ::P7Global] {
	set gt $::P7Global($g)
	if { $gt eq "tcl" } {
	    puts "  p7tcl_Unlink ( world->globals->[identifize g_$g] );"
	} elseif [is-refc-type $gt] {
	    puts "  p7obj_Unlink ( (p7obj) world->globals->[identifize g_$g] );"
	}
    }
    puts "  memset( world->globals, 0, sizeof *world->globals );"
    puts "\}"
}
proc master-p7class {name memberlist} {
    define-p7class $name $memberlist 
}
proc define-p7class {s memberlist} {

puts stderr "-- p7class $s    {$memberlist}"
    check "p7class name $s reserved for primative" not info exists ::Prim($s) ;# dont override primatives

    if {[info exists ::Struct($s)] && [llength $::Struct($s)]} {
	throw "redefining p7class $s"
    }

    # creating empty record marks it incomplete
    set ::Struct($s) {}
    set is [mangle $s]

    #TODO
	# declaring incomplete p7class

    foreach mem $memberlist {
	parse-typed-id mt m $mem
	    lappend ::Struct($s) $m
	    set ::MemType($s.$m) $mt
    }
    
    # add bogus record if p7class has no members
    if {![ll $::Struct($s)]} {
	lappend ::Struct($s) _emptiness_
	set ::MemType($s._emptiness_) int
    }

    # generate C code

    puts "#define ${is}_Link(p)	 p7obj_Link((p7obj)(p))"
    puts "#define ${is}_Unlink(p)	 p7obj_Unlink((p7obj)(p))"
    puts "#define ${is}_Assign(i,rv,x) p7tup_Assign(i,(p7obj*)(rv),(p7obj)(x))"
    puts "#define p7tcl_to_$is\(i,rv,x\)  p7tcl_to_p7tup(i, (p7obj*)(rv),x, & p7class_$is )"
    puts "#define p7tcl_from_$is\(i,rv,x\)  p7tcl_from_p7tup(i, rv, (p7obj)x)"

    puts "typedef struct p7struct_$is *$is;"
    puts "struct p7struct_$is \{"

    # builtin fields
    puts "#ifdef P7_MAGIC"
    puts "  int magic;"
    puts "#endif"
    puts "  int ref_count;"
    puts "  struct struct_p7cls* cls;"

    foreach m $::Struct($s) {
	set mt $::MemType($s.$m)
	set mmt [mangle $mt]
	puts "  $mmt m_$m;"
    }
    puts "\};"

    puts "static void p7Cleanup_$is ( p7obj ptr ) \{"
    set emitted_p false
    foreach m $::Struct($s) {
	set mt $::MemType($s.$m)
	if { $mt eq "tcl" } {
	    if {!$emitted_p} { puts "  $is p= ($is) ptr;" ; set emitted_p true }
	    puts "  if (p->m_$m) p7tcl_Unlink( p->m_$m );"
	} elseif {[is-refc-type $mt]} {
	    if {!$emitted_p} { puts "  $is p= ($is) ptr;" ; set emitted_p true }
	    puts "  if (p->m_$m) p7obj_Unlink( (p7obj) p->m_$m );"
	}
    }
    puts "  return;"
    puts "\}"
    puts "struct struct_p7cls p7class_$is = { "
    puts "#ifdef P7_MAGIC"
    puts "    P7_MAGIC,"
    puts "#endif"
    puts "    0x40000000,"
    puts "    NULL,"
    puts "    \"$s\","
    puts "    [set ::ClsNum($s) [incr ::NextClsNum]],"
    puts "    sizeof(struct p7struct_$is),"
    puts "    p7Cleanup_$is"
    puts "};"
}

proc master-p7proc {name paramlist body} {
    append body " " ;# workaround a lexxing bug at EOF
    declare-p7proc $name $paramlist
    lappend ::DefineP7ProcList $name $paramlist $body
}
proc declare-p7proc {namespec paramlist} {
    
#puts stderr [human "--declaring $namespec {$paramlist}"]

    new-cfunc

    # get return type

    parse-typed-id rt sn $namespec
    set ::ShortName $sn
    set ::Rtype $rt
    set mrt [mangle $rt]

    set ::Where "    in p7proc $sn"

    # get formal params
    foreach pspec $paramlist {
	parse-typed-id type name $pspec
	set name var_$name
	new-var $name $type "<PARAM>"
	lappend ::Param $name
    }

    set np [ll $::Param]
    set p0 [lindex $::Param 0]
    set t0 [expr {$np ? $::Type($p0) : "void" }]
    set mt0 [mangle $t0]
    lappend ::Counts($sn) $np 
    lappend ::Firsts($sn,$np) $t0
    set ::Pname [identifize "${sn}_${np}___$mt0"]
    lappend ::Pprocs $::Pname
    set ::Where "    in p7proc $sn {$::Param}   (* $::Pname *)"

    # remember signature
    lappend ::Sig($::Pname) $::Rtype
    foreach p $::Param { lappend ::Sig($::Pname) $::Type($p) }

    # dynamic must return same
    if { $np > 0 } {
	if { [is-dyna-type $t0] } {
	    lappend ::DynaFirsts($sn,$np) $t0
	    if {[ie ::DynaReturn($sn,$np)]} {
		check {conflicting dynamic return types for $sn with $np args: $rt vs $::DynaReturn($sn,$np)} eq $rt $::DynaReturn($sn,$np)
	    } else {
		set ::DynaReturn($sn,$np) $rt
	    }
	}
    }

    ##### Emit Forward Declaration
    lappend z "p7world world"
    if { $::Rtype == "void" } {
	# nothing
    } else {
	lappend z "[mangle $::Rtype] *retval"
    }
    foreach p $::Param {
	lappend z "[mangle $::Type($p)] $p"
    }
    puts "extern p7err p7func_$::Pname ( [join $z ,] );\n"

    unset ::Where
}

proc define-p7proc {namespec paramlist body} {
    ##########################################################
    new-cfunc

    # get return type

    parse-typed-id rt ::ShortName $namespec
    set ::Rtype $rt
    set mrt [mangle $rt]

    set ::Where "    in p7proc $::ShortName"
puts stderr [human "-- p7proc $namespec    {$paramlist}"]

    # get formal params
    foreach pspec $paramlist {
	parse-typed-id type name $pspec
	set name var_$name
	new-var $name $type "<PARAM>"
	lappend ::Param $name
    }

    set nparam [ll $::Param]
    set p0 [lindex $::Param 0]
    set t0 [expr {$nparam ? $::Type($p0) : "void" }]
    set mt0 [mangle $t0]
    set ::Pname [identifize "${::ShortName}_${nparam}___$mt0"]

    set ::Where "    in p7proc $::ShortName {$::Param}   (* $::Pname *)"
    ###########################################################

    dcl-type NULL null

    # get body of proc
    lex_initok $body

    set rv [p-stmts]

    if { $rt ne "void" && $rv ne "never" } {
	    set z [coerce "in result of p7proc $::ShortName (* $::Pname *)" $rt $rv]

	    code " ${mrt}_Assign ( world, retval , $z);"
	    code " goto final;"
    }

    unset ::Llinenum
    unset ::Ltokin
    unset ::Where
    emit-p7func
    emit-p7wrap
}


proc emit-p7func {} {
    set ::Where "    in p7proc %::Pname"
    # output C Function
    lappend z "p7world world"
    if { $::Rtype == "void" } {
	# nothing
    } else {
	lappend z "[mangle $::Rtype] *retval"
    }
    foreach p $::Param {
	lappend z "[mangle $::Type($p)] $p"
    }
    puts "\n/**********  $::Pname  **************/\n"
    puts "p7err p7func_$::Pname ( [join $z ,] )"
    puts "\{"
    puts "p7err err= TCL_OK;"
    puts "$::Init"
    puts "$::Link"
    puts "$::Code"
    puts "$::Final"
    puts "return err;"
    puts "\}"
    unset ::Where 
}

proc emit-megawrappers {} {
    foreach p [array names ::Counts] {
	emit-megawrapper $p
    }
}
proc emit-megawrapper {p} {
    new-cfunc
    # p7megawrappers are from TCL -- switches on number of args, to p7wrap function
    set ::Where "    in tcl-mega-wrapper for p7proc $::Pname"
    
    set ::Init "/*INIT*/\n"
    set ::Link "/*LINK*/\n"
    set ::Code "/*CODE*/\n"
    set ::Final "final: /*FINAL*/\n"

    new-var result tcl

    code "   switch ( objc-1 ) \{"
    set counts [lsort -unique -integer $::Counts($p)]
    foreach n $counts {
	if { $n==0 } {
	    code "      case 0 : err= p7wrap_[identifize ${p}_0___void] (cd, interp, objc, objv);"
	    code "      goto final;"
	} elseif {[ie ::DynaFirsts($p,$n)]} {
	    set tt [lrepeat tcl $n]
	    request-dyna $p $tt
	    set rt $::DynaReturn($p,$n)
	    set aa world
	    if { $rt ne "void" } {
		set rv retval_${n}
		new-var $rv $rt
		lappend aa &$rv
	    }
	    set i 0
	    foreach ty $tt {
		incr i
		lappend aa "objv\[$i\]"
	    }
	    code "      case $n : err= p7dyna_[identifize ${p}_$n][type-tail $tt] ( [join $aa "," ] );"
	    code "                if (err) goto final;"

	    if { $rt ne "void" } {
		code-assign result $rv
		code "	if (result) Tcl_SetObjResult( world->interp, result ); else Tcl_ResetResult( world->interp );"
	    } else {
		code "	Tcl_ResetResult( world->interp );"
	    }
	    code "      goto final;"

	} elseif {1==[set firsts [lsort $::Firsts($p,$n)]; ll $firsts]} { 
	    # singleton -- just call it

	    set ty [0 $firsts]
	    set mty [mangle $ty]

	    code "      case $n : err= p7wrap_[identifize ${p}_${n}___$mty] (cd, interp, objc, objv);"
	    code "      goto final;"
	} else {
	    # else error
	    code "      case $n : err= p7set_error_static( world, \"cannot call overloaded '$p' with $n args from Tcl (* options are '$firsts' *)\" );"
	    code "      goto final;"
	}


    }
    code "  \} /* end switch */"
    code "   err= p7set_error_d( world, \"no p7proc '$p' with %d args from Tcl (* options: $counts *)\", objc-1 );"
    code "   goto final;"

    # standard Object Command function header
    puts "int [identifize p7megawrap_$p] ( ClientData cd, Tcl_Interp* interp, int objc, Tcl_Obj *const* objv )"
    puts "\{"
    puts "   p7err err= TCL_OK;"
    puts "   p7world world= (p7world)cd;"

    puts "$::Init"
    puts "   assert(interp);"
    puts "   assert(world);"
    puts "$::Link"
    puts "$::Code"
    puts "$::Final"
    puts "return err;"

    puts "\}"
    unset ::Where 
}
proc request-dyna {meth types} {
    check "request-dyna must have first type tcl" eq [0 $types] "tcl"
    set ::RequestDyna([list $meth $types]) 1

    #####  Create the Forward Declartion for C
    set np [ll $types]
    set rt $::DynaReturn($meth,$np)
    set mrt [mangle $rt]
    lappend cparams "p7world world"
    if {$rt ne "void"} {
	lappend cparams "$mrt *retval"
    }
    set i 0
    foreach ty $types {
	lappend cparams "[mangle $ty] arg_[incr i]"
    }
    set typetail [type-tail $types]
    set dname [identifize "p7dyna_${meth}_${np}${typetail}"]

    ##### Append the declaration to ::Init
    append ::Init "extern p7err $dname ( [join $cparams ","] ); \n"

}
proc emit-dyna-all {} {
    foreach x [names ::RequestDyna] {
	eval emit-dyna $x
    }
}
proc type-tail {types} {
    set typetail ""
    foreach ty $types {
	append typetail ___[mangle $ty]
    }
    set typetail
}

proc emit-dyna {meth types} {
    new-cfunc

    set ::Init "/*INIT*/\n"
    set ::Link "/*LINK*/\n"
    set ::Code "/*CODE*/\n"
    set ::Final "final: /*FINAL*/\n"

    set np [ll $types]

    # what kind of return?
    set rt $::DynaReturn($meth,$np)
    set mrt [mangle $rt]
    lappend cparams "p7world world"
    if {$rt ne "void"} {
	lappend cparams "$mrt *retval"
    }

    # what kind of formal params?
    set i 0
    foreach ty $types {
	lappend cparams "[mangle $ty] arg_[incr i]"
    }
    set typetail [type-tail $types]


    code "  if ( arg_1 && arg_1->typePtr == &p7classObjType ) \{"
    code "     obj= (p7obj)arg_1->internalRep.otherValuePtr;"
    code "  \}"

    code "  if (obj) \{"
    code "     cls= obj->cls;"
    code "     if (cls) clsnum= cls->clsnum;"
    code "  \}"

    set firsts $::DynaFirsts($meth,$np)

    # if only one option, (tcl)"" can go there.
    if {[ll $firsts]==1} {
	set f [0 $firsts]
	if {$f ne "tcl"} {
	    # option is not a "tcl" option, convert empty string to the only clsnum
	    code "  if (arg_1 && !obj && Tcl_GetString(arg_1)\[0\]=='\\0' ) { clsnum= $::ClsNum($f); }"
	}
    }

    code "  switch (clsnum) \{"

    foreach f $firsts {
	set mf [mangle $f]

	if { $f eq "tcl" } {
	    code "  case 0 : "
	    code "  default : "
	} else {
	    code "  case $::ClsNum($f) : "
	}
	code "\{"

	set fid [identifize ${meth}_${np}___${mf}]
	set sig $::Sig($fid)

	set cargs "world"
	if {$rt ne "void"} {
	    # retval is already a pointer
	    lappend cargs "retval"
	}
	lappend cargs [expr { $f eq "tcl" ? "arg_1" : "($mf)obj" }]

	set i 1
	foreach ty [1..end $types] pt [2..end $sig] {
	    incr i
	    
#' ARG $i ty $ty pt $pt 
	    if { $ty eq $pt } {
		# good! same!
		lappend cargs arg_$i
	    } elseif { $ty eq "null" } {
		lappend cargs "([mangle $pt])NULL"
	    } else {
		#set mty [mangle $ty]
		#puts "  $mty $tmp= ($mty)0;"
		#puts "  err= p7tcl_to_$mty ( world, & $tmp, arg_$i );"
		#puts "  if (err) return err;"

		set tmp tmp_[serial]
		new-var $tmp $pt
		dcl-type arg_$i $ty
		code-assign $tmp arg_$i
		lappend cargs $tmp
	    } 
	}

	code "    err= p7func_$fid ( [join $cargs ","] );"

	code "    goto final;"
	code "  \}"
    }
    code "  \} /*end switch*/"
    
    if {![lhas $firsts "tcl"]}  {
	code "    err=  p7set_error_s(world, \"no dynamic p7proc '$meth' takes tcl, for $np args (* static types= [list $types] ; First Dynamic Type= %s *)\", cls? cls->clsname: \"tcl\" );"
	code "    goto final;"
    }
	    

    set dname [identifize "p7dyna_${meth}_${np}${typetail}"]
    puts "p7err $dname ( [join $cparams ","] ) \{"
    puts "  p7err err= TCL_OK;"
    puts "  int clsnum= 0;"
    puts "  p7cls cls= (p7cls)NULL;"
    puts "  p7obj obj= (p7obj)NULL;"
    puts "  p7tcl tcl= (p7tcl)NULL;"
    puts "  tcl= tcl;"

    puts "$::Init"
    puts "$::Link"

    puts "$::Code"
    puts "$::Final"
    puts "return err;"
    puts "\}"

}



proc emit-p7wrap {} {
    set ::Where "    in tcl-wrapper for p7proc $::Pname"
    set rt $::Rtype
    # standard Object Command function header
    puts "p7err p7wrap_$::Pname ( ClientData cd, Tcl_Interp* interp, int objc, Tcl_Obj *const* objv )"
    puts "\{"
    puts "  p7world world= (p7world)cd;"
    puts "  p7err err= TCL_OK;"
    if {$rt ne "void"} {puts "  p7tcl rettcl= (p7tcl)NULL;"}
    puts "  assert(world);"
    lappend z world
    set finals ""
    # prepare retval
    if { $rt == "void" } {
	# nothing
    } else {
	set mrt [mangle $rt]
	puts "  $mrt retval= ($mrt)0;"
	lappend z &retval
	if {[is-refc-type $rt]} { 
	    append finals "DOING(\"final unlink retval\");"
	    append finals "if (retval) { ${mrt}_Unlink(retval); retval= ($mrt)0; }\n"
	}
    }
    # prepare parameters
    foreach p $::Param {
	set pt $::Type($p)
	set mpt [mangle $pt]
	puts "  $mpt $p= ($mpt)0;"
	lappend z $p
	if {[is-refc-type $pt]} { 
	    append finals "DOING(\"final unlink $p\");"
	    append finals "if ($p) { ${mpt}_Unlink($p); $p= ($mpt)0; }\n"
	}
    }
    # convert params from Tcl
    set i 0
    foreach p $::Param {
	incr i
	set pt $::Type($p)
	set mpt [mangle $pt]
	puts "DOING(\"tcl param to $p\");"
	puts "  err= p7tcl_to_$mpt ( world, &$p, objv\[$i\] ); "
	puts "  [if-err-jump "converting parameter $i from tcl to type $pt"] "
    }

    # CALL THE FUNC
    puts "DOING(\"p7func $::Pname\");"
    puts "  err= p7func_$::Pname ( [join $z ,] );"
    puts "  [if-err-jump]"

    # convert retval to Tcl
    if { $rt ne "void" } {
	puts "  err= p7tcl_from_$mrt ( world, &rettcl, retval );"
	puts "  [if-err-jump "converting result (type $rt) to tcl"] "
	puts "  if (rettcl) Tcl_SetObjResult( interp, rettcl); else Tcl_ResetResult( interp );"

	append finals "DOING(\"final unlink rettcl\");"
	append finals "if (rettcl) { p7tcl_Unlink(rettcl); rettcl= (p7tcl)0; }\n"
    }
    
    puts "final:\n  $finals"
    #puts "CHECK"
    puts "return err;"
    puts "\}"
    unset ::Where 
}

proc is-stmt-ender {} {
    expr { $::Lt=="zz" || $::Lt=="rb" || $::Lt=="rs" || $::Lt=="sc" || $::Lt=="nl" }
}

proc p-block-eow {} {
    p-get "\{"
    set z [p-stmts]
    p-get "\}"
    end-of-word
    set z
}
proc p-stmts {} {
    set z void
    while { $::Lt!="rb" && $::Lt!="rs" && $::Lt!="zz" } {
        # skip null statements
        if { $::Lt == "sc" || $::Lt=="nl"  } {
	    lex_black
	    continue
        }
        if { $::Ld == "#" } { ;#comment
            while { $::Lt!="nl" } { lex_black }
	    continue
	}
	set z [p-stmt]
	check "expected end of statement, but got $::Ld" is-stmt-ender
    }
    return $z
}
proc p-stmt {} {

code "ATLINE( $::Llinenum );"

set human [human $::Ld$::Ltokin]
regsub {[\'\"\\]} $human {\\&} human
code "// BEGIN STATEMENT: $human"
code "CHECK"

    set pn ?
    catch {set pn $::Pname}
    catch {set line $::Llinenum}
    catch {set first $::Ld}
    catch {set rest $::Ltokin}
    within {within p7proc $pn, line $line, statement \"[string range [human "$first$rest"] 0 31]...\"} {

    if { $::Ld=="len" } {
	set z [p-len-stmt]
    } elseif { $::Ld=="p7peek" } {
	set z [p-p7peek-stmt]
    } elseif { $::Ld=="tclpeek" } {
	set z [p-tclpeek-stmt]
    } elseif { $::Ld=="lappend" } {
	set z [p-lappend-stmt]
    } elseif { $::Ld=="append" } {
	set z [p-append-stmt]
    } elseif { $::Ld=="new" } {
	set z [p-new-stmt]
    } elseif { $::Ld=="set" } {
	set z [p-set-stmt]
    } elseif { $::Ld=="incr" } {
	set z [p-incr-stmt]
    } elseif { $::Ld=="return" } {
	set z [p-return-stmt]
    } elseif { $::Ld=="error" } {
	set z [p-error-stmt]
    } elseif { $::Ld=="if" } {
	set z [p-if-stmt]
    } elseif { $::Ld=="while" } {
	set z [p-while-stmt]
    } elseif { $::Ld=="for" } {
	set z [p-for-stmt]
    } elseif { $::Ld=="loop" } {
	set z [p-loop-stmt]
    } elseif { $::Ld=="catch" } {
	set z [p-catch-stmt]
    } elseif { $::Ld=="break" } {
	set z [p-break-stmt]
    } elseif { $::Ld=="continue" } {
	set z [p-continue-stmt]
    } elseif { $::Ld=="k" } {
	set z [p-k-stmt]

    } elseif { [string match {::*} $::Ld] } {
	# initial :: means calling a command in Tcl
	set z [p-tcl-eval-stmt]

    } elseif { $::Lt=="id" && [info exists ::Counts($::Ld)] } {
	# calling a p7proc in p7
	set z [p-p7call-stmt]
    } else {
	throw "unknown statement beginning '$::Ld'"
    }

    }
code "// END STATEMENT: $human"
    return $z
}

proc p-k-stmt {} { # "k combinator" -- has value of its FIRST argument
    p-get "k"

    set ww {} ;# arguments to k

    while {![is-stmt-ender]} {
	set word [p-word]
    	lappend ww $word
    }

    #if { ! [llength $ww]} { error "k statement requires at least one argument" }

    if {[llength $ww]} {
	# return first arg
	return [0 $ww]
    } else {
        # k with no argument returns 0
        return NULL
    }

}
proc p-p7call-stmt {} {
    set pn $::Ld ;# procname
    lex_black

    set aa {} ;# arg list to pass
    set tt {} ;# arg list to pass
    while {![is-stmt-ender]} {
	set word [p-word]
    	lappend aa $word
    	lappend tt $::Type($word)
    }
    set nparam [ll $aa]
    set p0 [lindex $aa 0]
    set t0 [expr {$nparam ? [lindex $tt 0] : "void" }]
    # promote intlets to int
    switch -exact $t0 {
	uni {set t0 int}
	utf {set t0 int}
	byt {set t0 int}
    }
    set mt0 [mangle $t0]
    set lpn [identifize "${pn}_${nparam}___$mt0"]

    # test first param ($t0)$p0 for dyna
    set num_firsts [expr {[info exists ::Firsts($pn,$nparam)] ? [ll $::Firsts($pn,$nparam)] : 0}]

    if { $t0 eq "void" } {
	check "no p7proc $pn for $nparam args" info exists ::Sig($lpn)
	set rt [lindex $::Sig($lpn) 0] ;# ret typ
    } elseif { $t0 eq "tcl" && $num_firsts > 1 } {
	# dispatch on TCL is always dynamic
	request-dyna $pn $tt
	set rt $::DynaReturn($pn,$nparam)
    } elseif { $num_firsts == 1 && ![ie ::Sig($lpn)] } {
	# dispatch to singleton 
	set t0 [0 $::Firsts($pn,$nparam)]
	set mt0 [mangle $t0]
	set lpn [identifize "${pn}_${nparam}___$mt0"]
	set tname calltmp_1_[serial]
	new-var $tname $t0
	code-assign $tname $p0
	set p0 $tname
	set aa [concat [list $p0] [1..end $aa]]
	set rt [lindex $::Sig($lpn) 0] ;# ret typ
    } else {
	check "no p7proc $pn for $nparam args, beginning with $t0 (* $tt *)" info exists ::Sig($lpn)
	set rt [lindex $::Sig($lpn) 0] ;# ret typ
    }

    #notused# set mrt [mangle $rt]
    #notused# set ptt [lrange $::Sig($lpn) 1 end] ;# param types


    if { $rt eq "void" } {
	set zname void
	set pp [list world] ;# prepend for retval
    } else {
    	set zname p7call_${lpn}_[serial]
	new-var $zname $rt 
	set pp [list world "&$zname"] ;# prepend for retval
    }

    if { $t0 ne "void" } { lappend pp $p0 }
    set i 1
    foreach ai [1..end $aa] ti [1..end $tt] si [2..end $::Sig($lpn)] {
	incr i
	if { $si eq $ti } {
	    # good, already correct type
	    lappend pp $ai
	} else {
	    set tname calltmp_${i}_[serial]
	    new-var $tname $si
	    code-assign $tname $ai
	    lappend pp $tname
	}
    }


    # test for dyna
    if { $t0 eq "tcl" && $num_firsts>1 } {
	set funcname p7dyna_${pn}_${nparam}[type-tail $tt]
    } else {
	set funcname p7func_$lpn
    }

    code " err= $funcname ( [join $pp " , "] );"
    code " [if-err-jump]"
    return $zname
}

proc p-tclpeek-stmt {} {
    p-get "tclpeek"

    set x [p-word]
    check {tclpeek requires arg of type 'tcl', not type '$::Type($x)'} eq tcl $::Type($x)

    set result tclpeek_[serial]
    new-var $result tcl

    # null is empty
    code "  $result= p7tcl_tclpeek( $x ); /* returns already-linked p7tcl */"

    return $result
}
proc p-p7peek-stmt {} {
    p-get "p7peek"

    set x [p-word]
    check {p7peek requires arg of reference type, not type '$::Type($x)'} is-refc-type $::Type($x)

    set result p7peek_[serial]
    new-var $result tcl

    # null is empty
    if { $::Type($x) eq "tcl" } {
	code "  $result= p7tcl_tclpeek( $x ); /* returns already-linked p7tcl */"
    } else {
	code "  $result= p7tcl_p7peek( (p7obj)$x ); /* returns already-linked p7tcl */"
    }

    return $result
}


proc p-len-stmt {} {
    p-get "len"

    set x [p-word]
    set xt $::Type($x)
    set name len_[serial]
    new-var $name int 
    if { $xt eq "tcl" } {
	# null is empty
	code "  if ( $x ) {"
	code "     err= Tcl_ListObjLength( world->interp, $x, & $name);"
	code "     [if-err-jump]"
	code "  } else { $name= 0; }"
    } else {
	check "calling len on a non-vector (* $x $xt *)" is-vector-type $xt
	code " $name= ($x) ? ($x)->length : 0;"
    }
    return $name
}



proc p-new-stmt {} {
    p-get "new"

    # get var type

    if {$::Lt=="lp"} {
	# new style, without parens
	p-lvalue type name make_ref
	set mt [mangle $type]
    } else {
	# old style, without parens
	set type [p-typespec]
	set mt [mangle $type]
	set name new_[serial]
	new-var $name $type 
    }

    check "typespec missing in 'new' statement" ne $type "?"
    check "'new' statement cannot create type $type (* not a reference type *)" is-refc-type $type

    if {[is-vector-type $type]} {
	set len [p-word]
	check "length parameter in 'new' statement must be integer (* $::Type($len) $len *)" is-intlike $::Type($len)
	set maybe_len ", $len"
    } else {
	set maybe_len ""
    }

    if {[is-p7class-type $type]} {
	code "  err= p7tup_New( world, (p7obj*) &$name, &p7class_$mt );" 
	code "  [if-err-jump]"
	code "  p7tup_Link( (p7obj) $name);" 
    } else {
	code "  err= ${mt}_New( world, &$name $maybe_len );" 
	code "  [if-err-jump]"
	code "  p7obj_Link( (p7obj) $name);" 
    }

    set name

}

proc coerce-assign-tcl-tup {var vt x xt} {
    return "  err= p7tcl_from_[mangle $xt] ( world, & $var, $x ); [if-err-jump]"
}
proc coerce-assign-tup-tcl {var vt x xt} {
    return "  err= p7tcl_to_[mangle $vt] ( world, & $var, $x ); [if-err-jump]"
}


proc coerce-assign<tcl,int> {dest src} {
    return "  p7tcl_Assign( world, & $dest, Tcl_NewIntObj( $src ) );"
}
proc coerce-assign<int,tcl> {dest src} {
    return "  err= Tcl_GetIntFromObj( world->interp, $src, & $dest ); [if-err-jump]"
}

# Convert TCL to Char Arrays 
proc coerce-assign<uni*,tcl> {dest src} {
    return "  if ($dest) p7obj_Unlink((p7obj) $dest);
    $dest = (p7uni_VEC_)NULL;
    err= p7tcl_to_p7uni_VEC_( world, & $dest, $src ); [if-err-jump]"
}
proc coerce-assign<utf*,tcl> {dest src} {
    return "  if ($dest) p7obj_Unlink((p7obj) $dest);
    $dest = (p7utf_VEC_)NULL;
    err= p7tcl_to_p7utf_VEC_( world, & $dest, $src ); [if-err-jump]"
}
proc coerce-assign<byt*,tcl> {dest src} {
    return "  if ($dest) p7obj_Unlink((p7obj) $dest);
    $dest = (p7byt_VEC_)NULL;
    err= p7tcl_to_p7byt_VEC_( world, & $dest, $src ); [if-err-jump]"
}

# Convert Char Arrays to TCL
proc coerce-assign<tcl,uni*> {dest src} {
    return "  if ($dest) p7tcl_Unlink($dest); 
      $dest = ($src) ? Tcl_NewUnicodeObj( $src->data, $src->length ) : NULL; 
      p7tcl_Link($dest);  "
}
proc coerce-assign<tcl,utf*> {dest src} {
    return "  if ($dest) p7tcl_Unlink($dest); 
      $dest = ($src) ? Tcl_NewStringObj( (char*) $src->data, $src->length ) : NULL; 
      p7tcl_Link($dest);  "
}
proc coerce-assign<tcl,byt*> {dest src} {
    return "  if ($dest) p7tcl_Unlink($dest); 
      $dest = ($src) ? Tcl_NewByteArrayObj( (void*) $src->data, $src->length ) : NULL; 
      p7tcl_Link($dest);  "
}


proc castable-assign {dest src} {
    if { $dest == $src } { return 1 }

    set inttypes "int byt utf uni 0"
    if {[lhas $inttypes $dest] && [lhas $inttypes $src]} {
	return 1
    }
    return 0
}

proc code-assign {var x} {

	    set vt $::Type($var)
	    set mvt [mangle $vt]

	    set xt $::Type($x)

	    if { $vt eq "int" && $xt eq "int" } {
		code "    $var = $x;   /* assign int */"
	    } elseif { $xt eq "null" || $xt eq "void" || $xt eq "never" } {
		code "    ${mvt}_Assign ( world, & $var, ($mvt)0 );"
	    } elseif {[ castable-assign $vt $xt]} {
		code "    ${mvt}_Assign ( world, & $var, ($mvt)$x );"
	    } elseif {[info command coerce-assign<$vt,$xt>] == "coerce-assign<$vt,$xt>"} {
		code [coerce-assign<$vt,$xt> $var $x]

	    } elseif { $vt eq "tcl" && [is-intlike $xt] } {
		code "  err= p7tcl_from_[mangle $xt] ( world, & $var, $x ); [if-err-jump]; "
	    } elseif { $xt eq "tcl" && [is-intlike $vt] } {
		code "  err= p7tcl_to_[mangle $vt] ( world, & $var, $x ); [if-err-jump]; "

	    } elseif { $vt eq "tcl" && [is-vector-type $xt] } {
		code [coerce-assign-tcl-tup $var $vt $x $xt]
	    } elseif { $xt eq "tcl" && [is-vector-type $vt] } {
		code [coerce-assign-tup-tcl $var $vt $x $xt]

	    } elseif { $vt eq "tcl" && [is-p7class-type $xt] } {
		code [coerce-assign-tcl-tup $var $vt $x $xt]
	    } elseif { $xt eq "tcl" && [is-p7class-type $vt] } {
		code [coerce-assign-tup-tcl $var $vt $x $xt]
	    } else {
		check "Cannot assign $vt from $xt -- perhaps not yet implemented?" expr 0
	    }
}

proc p-lvalue {typeName varName refcodeName} {
    upvar 1 $typeName type $varName name $refcodeName make_ref
    # get var type
    set type [p-opt-paren-typespec]

    if { $type eq "?" } {
	# no type spec -- can be arbitrary lvalue
	# capture the reference code -- use it after the value is found
	push-code
	set name [p-ref-var 1]
	set make_ref [pop-code]


    } else {
	# there was a typespec -- must be single identifier, naming a new variable
	set mt [mangle $type]

	if { $::Lt=="id" && $::Lwhite=="" } {
	    # get var name
	    set name var_[p-get-ident]
	} else {
	    set name anon_[serial]
	}
	set make_ref ""

	# register new var
	new-var $name $type 
    }
    within {end-of-word required after lvalue spec, but got '$::Ld' (* ($type) $name *)} end-of-word
}

proc p-set-stmt {} {
    p-get "set"
    p-lvalue type name make_ref

    if {[is-stmt-ender]} {
	    # really a get, not a set

	    code $make_ref

    } else {

	    set y [p-word]

	    # important to emit $make_ref right before assign (or it could point
	    #   to freed/reallocated memory

	    code $make_ref

	    check "not allowed to assign to variable this" ne $name var_this
	    code-assign $name $y

	    if {[sm var_* $name]} {
		# return void if the command sets.   this is for performance.
		return void
	    }
    }

    # make new safe var for the result
    set namet $::Type($name)
    set mnamet [mangle $namet]
    set z set_[serial]
    new-var $z $namet 
    code-assign $z $name
    set z
}
proc p-incr-stmt {} {
    p-get "incr"
    p-lvalue type name make_ref

    if {[is-stmt-ender]} {
	set x 1   ;# default increment is 1
	dcl-type 1 int
    } else {

	    set x [intize [p-word]]
    }

    set nt $::Type($name)
    set xt $::Type($x)
    check "non-int to be incremented by incr (* $nt $name *)" eq int $nt 
    check "non-int increment in incr statement (* $xt $x *)" eq int $xt 
    check "not allowed to assign to variable this" ne $name var_this

    # make new safe var for the result
    set mnt [mangle $nt]
    set z incr_[serial]
    new-var $z $nt 

    code $make_ref
    code "  $z= ($mnt) ( $name += $x );"
    set z
}

proc p-append-stmt {} {
    p-get "append"

    # var to modify
    p-lvalue type name make_ref

    # x will be added
    set x [tclize [p-word]]

    # check types
    set nt $::Type($name)
    set xt $::Type($x)
    check "non-tcl to be incremented by incr (* $nt $name *)" eq tcl $nt 
    check "not allowed to assign to variable this" ne $name var_this

    # do the deed
    code "   if (! $name ) { $name= p7tcl_new_linked(); }"

    code "   if (Tcl_IsShared($name)) { p7tcl_Unlink( $name ); $name= Tcl_DuplicateObj( $name ); p7tcl_Link( $name ); }"
    code "   Tcl_AppendObjToObj( $name, $x );"


    # make new safe var for the result
    set mnt [mangle $nt]
    set z lappend_[serial]
    new-var $z $nt 

    code $make_ref
#void    code-assign $z $name
#void    set z
    return void
}

proc p-lappend-stmt {} {
    p-get "lappend"

    # var to modify
    p-lvalue type name make_ref

    # x will be added
    set x [tclize [p-word]]

    # check types
    set nt $::Type($name)
    set xt $::Type($x)
    check "non-tcl to be incremented by incr (* $nt $name *)" eq tcl $nt 
    check "not allowed to assign to variable this" ne $name var_this

    # do the deed
    code "  if (! $name ) { $name= p7tcl_new_linked(); }"

    code "  if (Tcl_IsShared($name)) { p7tcl_Unlink( $name ); $name= Tcl_DuplicateObj( $name ); p7tcl_Link( $name ); }"
    code "  err= Tcl_ListObjAppendElement( world->interp, $name, $x ); [if-err-jump]"

#void    # make new safe var for the result
#void    set mnt [mangle $nt]
#void    set z lappend_[serial]
#void    new-var $z $nt 

    code $make_ref
#void    code-assign $z $name
#void    set z
    return void
}

proc coerce {msg required x} {
    if {$x eq "void"} {throw "cannot coerce void to type $required (* $x *)"}
    set xt $::Type($x)
    if {$required eq "void"} {throw "cannot coerce type ($xt) to void (* $x *)"}
    if { $required eq $xt } {
	return $x
    } elseif {[is-intlike $required] && [is-intlike $xt]} {
	set mreq [mangle $required]
	set z "(($mreq)($x))"
	set ::Type($z) $required
	return $z
    } else {
	set mreq [mangle $required]
	set z coerce_${mreq}_[serial] 
	new-var $z $required 
	catch {set msg [uplevel 1 [list subst $msg]] }
	check {$msg: cannot convert $xt to $required (* $x *) (* $what *)} not catch { code-assign $z $x } what
	return $z
    }
}

proc p-return-stmt {} {
    p-get "return"

    if { [is-stmt-ender] } {
	# void return
	check "no-arg 'return' statement returns void, but should return $rt" eq void $rt
	code " /* --- RETURN VOID --- */ "
	code " goto final;"
    } else {
	# expression return
	set x [p-word]
	set xt $::Type($x)
	set rt $::Rtype
	set mrt [mangle $rt]

	set z [coerce {'return' statement returns $xt, but should return $rt} $rt $x]

	code " ${mrt}_Assign ( world, retval , $z);"
	code " /* --- RETURN --- */ "
	code " goto final;"
    }
    return never
}

proc p-error-stmt {} {
    p-get "error"

    # expression error
    set x [tclize [p-word]]

    code "  if ($x) Tcl_SetObjResult( world->interp, $x); else Tcl_ResetResult( world->interp );"
    code "  err= TCL_ERROR;"
    
    code " /* --- THROW ERROR --- */ "
    code [if-err-jump [list "in p7 'error' statement"]]
    return never
}


proc p-tcl-eval-stmt {} {

    set cmd tcleval_[serial]
    new-var $cmd tcl 

    dcl-type "Tcl_NewObj()" tcl
    code-assign $cmd "Tcl_NewObj()"

    set tmp word_[serial]
    new-var $tmp tcl 

    #code "Tcl_ResetResult(world->interp);"

    set i 0
    while { ! [is-stmt-ender] } {
        incr i
code "/* $cmd Begin Arg $i */ ;"
	set w [p-word]
	set wt $::Type($w)
	set mwt [mangle $wt]

code "/* begin code-assign tmp= $tmp w= ($wt) $w */"
	code-assign $tmp $w
code "/* end code-assign tmp= $tmp w= ($wt) $w */"

	code "  err= Tcl_ListObjAppendElement(world->interp, $cmd, $tmp? $tmp: Tcl_NewObj() );"
	code "  [if-err-jump]"

code "/* $cmd End Arg $i --> $mwt $w */ ;"
    }

code "/* $cmd Eval */"
    code "  err= Tcl_EvalObjEx(world->interp, $cmd, TCL_EVAL_DIRECT);"
    code "  [if-err-jump]"

    dcl-type "Tcl_GetObjResult(world->interp)" tcl
    code-assign $tmp "Tcl_GetObjResult(world->interp)"

    return $tmp
}

proc intize {x} {
    set xt $::Type($x)
    if {[is-intlike $xt]} {
	dcl-type "($x)" int
	return "($x)"
    } elseif { $xt == "tcl" } {
	set name intize_[serial]
	new-var $name int 
	code "  err= p7tcl_to_p7int (world, &$name, $x);"
	code "  [if-err-jump]"
	return $name
    } else {
	error "cannot use type $xt as a integer (* $x *)"
    }
}
proc conditionize {x} {

    set xt $::Type($x)
    if { $xt == "null" } {
	dcl-type 0 int
	return 0
    } elseif {[is-intlike $xt]} {
	dcl-type (0!=$x) int
	return (0!=$x)
    } elseif { $xt == "tcl" } {
	set name conditionize_[serial]
	new-var $name int 
	code "  err= p7tclcond(world, &$name, $x);"
	code "  [if-err-jump]"
	return $name
    } elseif {[is-vector-type $xt]} {
	dcl-type (NULL!=$x) int
	return (NULL!=$x)
    } elseif {[is-p7class-type $xt]} {
	dcl-type (NULL!=$x) int
	return (NULL!=$x)
    } else {
	error "should not default in conditionize (* $x $xt *)"
    }
}
	
	


proc p-if-stmt {} {
    p-get "if"
    set result never ;# tentatively, optimisticly

    set cond [conditionize [p-word]]

    if { $::Ld=="then" } { p-get then }
    code "if ( $cond ) \{"

	set z [p-block-eow]
	if { $z!="never" } { set result void }

    code "\}"
    set count 0
    while { $::Ld=="elseif" } {
	code "else \{"
	incr count
	p-get elseif
	set cond [conditionize [p-word]]

	if { $::Ld=="then" } { p-get then }
	
	code "if ( $cond ) \{"

	set z [p-block-eow]
	if { $z!="never" } { set result void }

	code "\}"
	
    }
    if { $::Ld=="else" || $::Lt=="lb" } {
	if { $::Ld=="else" } { p-get else }
	code "else \{"

	set z [p-block-eow]
	if { $z!="never" } { set result void }

	code "\}"
    }
    for {set i 0} {$i<$count} {incr i} {
	code "\}"
    }
    return $result;
}

proc p-loop-stmt {} {
    push-var ::ContinueLabel continue_[serial]
    p-get "loop"
    p-lvalue type indx make_ref

    set termination [intize [p-word]]

    code "  $make_ref"
    code "  $indx= 0;"
    code "  while (  $indx < $termination ) \{"

    p-block-eow

    if {[info exists ::LabelUsed($::ContinueLabel)]} {code "$::ContinueLabel: ;"}

    code "  $make_ref"
    code "  ++ $indx;"
    code "  \}"
    pop-var ::ContinueLabel 
    return void

}

proc p-catch-stmt {} {

    p-get "catch"

    push-var ::FinalLabel catcher_[serial]
    set x [p-block-eow]
    code "  goto $::FinalLabel;"
    code "$::FinalLabel: "
    pop-var ::FinalLabel

    set z catch_[serial]
    new-var $z int
    code "  $z= err;"
    code "  err= TCL_OK;"


    set r catch_result_[serial]
    new-var $r tcl
    code "  if ($r) p7tcl_Unlink( $r );"
    code "  $r= (p7tcl)NULL;"
    code "if ( $z ) \{ "
	code "  $r= Tcl_GetObjResult(world->interp);" 
	code "  p7tcl_Link( $r );"
    code "\} else \{"
	code-assign $r $x
    code "\}"

    set errInfoLit "world->literals\[ $::LIT_ERRORINFO \]"
    set errCodeLit "world->literals\[ $::LIT_ERRORCODE \]"

    code "  if (world->globals->g_p7errorInfo) p7tcl_Unlink (world->globals->g_p7errorInfo);"
    code "  world->globals->g_p7errorInfo= Tcl_ObjGetVar2(world->interp, $errInfoLit, NULL, TCL_GLOBAL_ONLY);"
    code "  if (world->globals->g_p7errorInfo) p7tcl_Link (world->globals->g_p7errorInfo);"

    code "  if (world->globals->g_p7errorCode) p7tcl_Unlink (world->globals->g_p7errorCode);"
    code "  world->globals->g_p7errorCode= Tcl_ObjGetVar2(world->interp, $errCodeLit, NULL, TCL_GLOBAL_ONLY);"
    code "  if (world->globals->g_p7errorCode) p7tcl_Link (world->globals->g_p7errorCode);"

    code "  Tcl_ResetResult(world->interp);"

    if {! [is-stmt-ender]} {
	# optional variable spec to put the error message / valid result $r in
	p-lvalue type name make_ref
	code "  $make_ref"
	code-assign $name $r
    }

    return $z
}

proc p-while-stmt {} {
    push-var ::ContinueLabel continue_[serial]
    p-get "while"

    push-code
    set cond [conditionize [p-word]]
    set make_cond [pop-code]

    code $make_cond
    code "while ( $cond ) \{"
    p-block-eow

    if {[info exists ::LabelUsed($::ContinueLabel)]} {code "$::ContinueLabel: ;"}
    code $make_cond
    code "\}"
    pop-var ::ContinueLabel 
    return void
}

proc p-for-stmt {} {
    push-var ::ContinueLabel continue_[serial]
    p-get "for"

    # initial block
    p-block-eow

    # condition
    push-code
    set cond [conditionize [p-word]]
    set make_cond [pop-code]
    
    #end-of-word

    # advance
    push-code
    p-block-eow
    set advance [pop-code]
    #end-of-word

    code $make_cond
    code "while ( $cond ) \{"
    p-block-eow
    #end-of-word

    if {[info exists ::LabelUsed($::ContinueLabel)]} {code "$::ContinueLabel: ;"}

    code $advance
    code $make_cond
    code "\}"
    pop-var ::ContinueLabel 
    return void
}

proc p-break-stmt {} {
    p-get "break"
    code "break;"
    return never
}

proc p-continue-stmt {} {
    p-get "continue"
    code "goto $::ContinueLabel;"
    set ::LabelUsed($::ContinueLabel) 1
    return never
}

proc new-literal {string} {
    set var "world->literals\[[incr ::NextLiteral]\]"
    dcl-type $var tcl
    code "  if ( ! $var ) \{"
    code "      $var= Tcl_NewStringObj( \"$string\", -1 );"
    code "      p7tcl_Link( $var );"
    code "  \}"
    set var
}

proc tcl-add-x {tclName x} {
    upvar 1 $tclName tcl
    if { $tcl eq "" } {
	set tcl $x
	###### i dont think the following was necessary
	###? set tcl dq_[serial]
	###? new-var $tcl tcl
	###? code-assign $tcl $x
    } else {
	set tcl [tclize $tcl]
	set x [tclize $x]
	code "  Tcl_AppendObjToObj( $tcl, $x ? $x : Tcl_NewObj() );"
    }
}
proc tcl-add-accum {tclName accumName final} {
    upvar 1 $tclName tcl $accumName accum
    if { $tcl eq "" } {
	if $final {
	    set tcl [new-literal $accum]
	} elseif { $accum ne "" } {
	    set tcl dq_[serial]
	    new-var $tcl tcl
	    code "  if ( $tcl ) p7tcl_Unlink( $tcl );"
	    code "  $tcl= Tcl_NewStringObj( \"$accum\", -1 );"
	    code "  p7tcl_Link( $tcl );"
	}
    } elseif { $accum ne "" } {
	set tcl [tclize $tcl]
	code "  if (Tcl_IsShared($tcl)) $tcl= Tcl_DuplicateObj($tcl);"
	code "  Tcl_AppendToObj( $tcl, \"$accum\", -1 );"
    }
    set accum ""
}
proc p-bareword {} {

	set tcl ""  ;# be lazy about allocking result
        set accum ""
        while 1 {
#' BARE '$::Lt' = '$::Ld'
                switch $::Lt {
                        zz { error "missing closing double-quote"}
                        ls {  ;# \[
			    tcl-add-accum tcl accum 0
			    set x [p-square]
			    tcl-add-x tcl $x
			}
			do {  ;# $
			    tcl-add-accum tcl accum 0
			    set x [p-deref-var 2]
			    tcl-add-x tcl $x
			}
			lp - rp - lb - dq {  ;# ( ) \{ \"
			    error "character '$::Ld' not allowed in middle of tcl word"
			}
			nl - sc - rb - rs {  ;# \n \; \} \]
			    break 
			}
                        default { append accum $::Ld ; lex_black }
                }
		if {[sl $::Lwhite]} break
#' BARE continue
        }
	tcl-add-accum tcl accum 1
#' BARE Break

	set tcl
}
proc p-doublequoted {} {
        lex_black  ;# advance past the dq

	set tcl ""  ;# be lazy about allocking result
        set accum ""
        while 1 {
		append accum $::Lwhite
                switch $::Lt {
                        zz { error "missing closing double-quote"}
                        dq { break }
                        ls {  ;# \[
			    tcl-add-accum tcl accum 0
			    set x [p-square]
			    tcl-add-x tcl $x
			}
			do {  ;# $
			    tcl-add-accum tcl accum 0
			    set x [p-deref-var 2]
			    tcl-add-x tcl $x
			}
                        default { append accum $::Ld ; lex_black }
                }
        }
	tcl-add-accum tcl accum 1
        p-get "\""

	set tcl
}
proc p-braced {} {
    p-get "\{"
    set buf ""
    set depth 1
    while { true } {
	if { $::Lt=="rb" } { 
	    append buf $::Lwhite
	    incr depth -1
	    if { $depth==0 } break
	    append buf $::Ld
	} elseif { $::Lt=="zz" } {
	    throw "missing close brace"
	} elseif { $::Lt=="lb" } {
	    append buf $::Lwhite $::Ld
	    incr depth
	} else {
	    append buf $::Lwhite $::Ld
	}
	lex_black
    }
    p-get "\}"

    regsub -all {[\'\"\\]} $buf {\\&} buf

    set var [new-literal $buf]

    # TODO subst with bareword
    set var
}

proc p-word {} {
#' p-word (* $::Lt $::Ld *)

    if { $::Lt == "lp" } {
	set z [p-pexpr]
    } elseif { $::Lt == "ls" } {
	#set z [p-square]
	set z [p-bareword]
    } elseif { $::Lt == "lb" } {
	set z [p-braced]
    } elseif { $::Lt == "dq" } {
	set z [p-doublequoted]
    } elseif { $::Ld == "\$" } {
	#set z [p-deref-var 1]
	set z [p-bareword]
    } elseif { $::Lt == "nu" } {
	set z [p-number]
    } elseif { $::Ld == "null" } { ;# keyword null
	set z [p-null]
    } elseif { $::Lt == "id" } { ;# any identifier
	set z [p-bareword]
    } elseif { $::Lt == "ch" } {
	set z [p-bareword]
    } else {
	throw "expected word, got '$::Ld' (* token type '$::Lt' *)"
    }
}

proc p-pexpr {} { ;# Parenthesized ParentheTcl Expression
    if { $::Lt == "lp" } {
	p-get "("
	set z [p-topexpr]
	p-get ")"
    } elseif { $::Lt == "ls" } {
	p-get "\["
	set z [p-topexpr]
	p-get "\]"
    } else {
	error "internal: funny start to pexpr: $::Ld"
    }
    set z
}

proc p-topexpr {} {
    p-condexpr
    #p-disjunc
}

proc p-condexpr {} {
    set x [p-disjunc]
    set z $x ;#tentative, if no ?

    if { $::Ld == "?" } {
	lex_black
	set x [conditionize $x]

	push-code
	set a [p-condexpr]
	set acode [pop-code]

	if { $::Ld != ":" } { error "Expected : after ? but got $::Ld" }
	lex_black
	push-code
	set b [p-condexpr]
	set bcode [pop-code]

	if { $::Type($a) != $::Type($b) } {
	    error "The type after `?' is `$::Type($a)' but the type after `:' is `$::Type($b)' -- they should be the same"
	}
	set zt $::Type($a) 

	set z ternary_[serial]
	new-var $z $zt

	code "if ($x) \{ /*ternary*/"
	code $acode
	code-assign $z $a
	code "\} else \{ /*ternary*/"
	code $bcode
	code-assign $z $b
	code "\} /*ternary*/"

    }

    set z
}


proc p-disjunc {} {
    set x [p-conjunc]

    while { $::Ld == "||" } {
	lex_black
	set x [conditionize $x]

        if {![info exists z]} {
		set z disj_[serial]
		new-var $z int 
		code "$z= $x;"
	}

	code "if (! $z) \{"

	set x [conditionize [p-conjunc]]

	code "   $z= $x;"
	code "\}"

    }

    expr { [info exists z] ? $z : $x }
}


proc p-conjunc {} {
    set x [p-relop]

    while { $::Ld == "&&" } {
	lex_black
	set x [conditionize $x]

        if {![info exists z]} {
		set z conj_[serial]
		new-var $z int "int $z=1;" 
		code "$z= $x;"
	}

	code "if ($z) \{"

	set x [conditionize [p-relop]]

	code "   $z= $x;"
	code "\}"

    }

    expr { [info exists z] ? $z : $x }
}

array set OpLevel { * 3 / 3 % 3 + 4 - 4 << 5 >> 5 >>> 5 < 6 <= 6 > 6 >= 6 lt 6 le 6 gt 6 ge 6 eq 7 ne 7 == 7 != 7 & 8 ^ 9 | 10 }


proc genx-tcl {msg list} {
    set zz [genx-tree $list]
    # TODO convert to tcl
    check {$msg (* [0 $zz] *)} eq tcl [1 $zz]
    0 $zz
}
proc genx-int {msg list} {
    set zz [genx-tree $list]
    set x [0 $zz]
    set xt [1 $zz]
    check {[uplevel [list subst $msg]] (* [0 $zz] *)} is-intlike $xt
    if { $xt eq "int" } {
	return $x
    } else {
	return "( (int)$x )"
    }
}
    
proc is-intlike {type} { 
    # true for int & intlets
    expr { $type eq "int" || $type eq "uni" || $type eq "utf" || $type eq "byt" }
}
proc untclize {type old} {
    # return value as a Tcl variable (if not already)
    set ot $::Type($old) 
    if { $ot eq "$type" } {
	return $old
    } else {
	set mt [mangle $type]
	set v untclize_${mt}_[serial]
	new-var $v $type 
	code-assign $v $old
	return $v
    }
}
proc tclize old {
    # return value as a Tcl variable (if not already)
    if { $::Type($old) eq "tcl" } {
	#dcl-type "p7tcl_newifnull($old)" tcl
	#return "p7tcl_newifnull($old)"
	check {NOT_VAR_IN_TCLIZE: $old} is-var $old
	code "  if (! $old) { $old= p7tcl_new_linked(); }"
	return $old
    } else {
	set v tclize_[serial]
	new-var $v tcl 
	code-assign $v $old
	return $v
    }
}
proc genx-tree list {
    # return a pair { c_code type } for the given tree
  set op [0 $list]
  switch -exact -- $op {
    factor {
	return [list [2 $list] [1 $list]]
    }
    * - / - % - + - - - << - >> - & - ^ - | {   
	    set l [genx-int {operator $op needs int on left} [1 $list]]
	    set r [genx-int {operator $op needs int on right} [2 $list]]
	    return [list "( $l $op $r )" int]
	}
    >>> {
	    set l [genx-int {operator $op needs int on left} [1 $list]]
	    set r [genx-int {operator $op needs int on right} [2 $list]]
	    return [list "p7unsigned_shift_right( $l , $r )" int]
	}
    == - != - < - <= - > - >= {
	    set ll [genx-tree [1 $list]]
	    set lx [0 $ll]
	    set lt [1 $ll]
	    set rr [genx-tree [2 $list]]
	    set rx [0 $rr]
	    set rt [1 $rr]
	    if { $lt eq "tcl" || $rt eq "tcl" } {
		check "With (tcl) data types, you must use operators eq,ne,lt,le,gt,ge instead of operators ==,!=,<,<=,>,>=" expr false
	    }
	    if { $lt ne $rt } {
		if { [is-intlike $lt] && [is-intlike $rt] } {
		    set lx "((p7int)($lx))"
		    set rx "((p7int)($rx))"
		} elseif { $lt eq "null" } {
		    set lx " ([mangle $rt])0 "
		} elseif { $rt eq "null" } {
		    set rx " ([mangle $lt])0 "
		} else {
		    check "Bad types: ($lt) operator $op ($rt)  (* $lx $op $rx *)" expr false
		}
	    }
	    return [list "($lx $op $rx)" int]
	}
    eq - ne - lt - le - gt - ge {
	    set ll [genx-tree [1 $list]]
	    set lx [0 $ll]
	    set lt [1 $ll]
	    set rr [genx-tree [2 $list]]
	    set rx [0 $rr]
	    set rt [1 $rr]
	    set lx [tclize $lx]
	    set rx [tclize $rx]
	    return [list "p7$op ($lx , $rx)" int]
	}
    }
    check "unknown operator $op" expr false
}

proc genx-arith {ll max} {
#' list genx-arith < $ll > max $max
    while {[info exists ::OpLevel($::Ld)]} {
	set op $::Ld
	set oplevel $::OpLevel($op)
	if { $oplevel >= $max } {
#' list RETURNING nextop $op $oplevel >= $max
	    return $ll
	    return [genx-arith $ll 999]
	}
#' list CONTINUING nextop $op $oplevel < $max
	p-get $op
	#set max $oplevel ;# now we dont want to go greater than what we've seen
	set seed [genx-basic]
	set rr [genx-arith $seed $oplevel]
	set ll [list $op $ll $rr]
    }
#' list ENDING ll $ll
    set ll
}
proc genx-basic {} {
    set z [p-factor]
    list factor $::Type($z) $z
}

proc p-relop {} {
    set seed [genx-basic]
    set tree [genx-arith $seed 999]
    set cc [genx-tree $tree]
    set z [0 $cc]
    set ::Type($z) [1 $cc]
    set z
}

proc p-factor {} {
    if { $::Lt == "nu" } { ;# number
    	p-number
    } elseif { $::Lt=="ls" } { ;# left-square \[
	p-square
    } elseif { $::Lt=="lp" } { ;# left-paren \( -- ParentheTcl Expr
	p-pexpr
    } elseif { $::Ld=="'" } { ;# apostrophe
	p-apostrophe
    } elseif { $::Lt=="dq" } { ;# double quote"
	p-doublequoted
    } elseif { $::Ld=="\$" } { ;# dollar $
	p-deref-var 0
    } elseif { $::Ld=="!" } { ;# monad !
	p-monad-bang
    } elseif { $::Ld=="~" } { ;# monad ~
	p-monad-snake
    } elseif { $::Ld=="-" } { ;# monad -
	p-monad-minus
    } elseif { $::Ld=="null" } { ;# keyword null
	p-null
    } elseif { $::Lt=="id" } { ;# id WITHOUT dollar -- clike
	p-deref-var 0
    } else {
    	throw "not a factor: ($::Lt) $::Ld"
    }
}

proc p-monad-bang {} {
    p-get "!"
    set f [conditionize [p-factor]]
    set z "!($f)"
    dcl-type $z int
    set z
}

proc p-monad-snake {} {
    p-get "~"
    set f [intize [p-factor]]
    set z "~($f)"
    dcl-type $z int
    set z
}

proc p-monad-minus {} {
    p-get "-"
    set f [intize [p-factor]]
    set z "-($f)"
    dcl-type $z int
    set z
}

proc p-square {} {
    p-get "\["
    set x [p-stmts]
    p-get "\]"
    set x
}

proc lookup-var {v} {

    set name var_$::Ld
    if { ! [is-var $name]} {
	if {[is-var var_this] && [info exists ::MemType($::Type(var_this).$::Ld)]} {
	    # TODO: safe if we forbid assigning to this
	    set name var_this->m_$::Ld
	    dcl-type $name $::MemType($::Type(var_this).$::Ld)
	} elseif {[ie ::P7Global($v)]} {
	    set gt $::P7Global($v)
	    set name "world->globals->g_$v"
	    dcl-type $name $gt
	} else {
	    check "variable $::Ld does not exist" expr 0
	}
    }
    set name
}

proc p-ref-var { bare } {

    if { $::Lt == "lb" } {
	# name in braces
	p-get "\{"
	check "variable name not valid identifier (* $::Ld *)" eq $::Lt id
	set name [lookup-var $::Ld]
	lex_black
	p-get "\}"
    } else {
	check "variable name not valid identifier (* $::Ld *)" eq $::Lt id
	set name [lookup-var $::Ld]
	lex_black
    }

    if {$bare} {

	while { $::Lwhite eq "" && ( $::Lt eq "lp" || $::Ld eq "." || $::Ld eq "->" )} {
	    set type $::Type($name)
	    if {[is-p7class-type $type]} {
		set name [p-subfield $name]
	    } else {
		set name [p-subscript $name]
	    }
	}

    } else {

	while { $::Lt eq "lp" || $::Lt eq "ls" || $::Ld eq "." || $::Ld eq "->" } { ;# left-paren subscript OR .fieldname
	    set type $::Type($name)
	    if {[is-p7class-type $type]} {
		set name [p-subfield $name]
	    } else {
		set name [p-subscript $name]
	    }
	}
    }

    set name
}
proc p-deref-var { bare } {
    if { $::Ld == "\$" } {
	p-get "\$"
    }

    set x [p-ref-var $bare]

# order of eval is wrong, if you dont make a new var :( see test_order

    set xt $::Type($x)
    set z deref_[serial]
    new-var $z $xt 
    code-assign $z $x
  
    set z
}
    
proc p-apostrophe {} {
    lex_getchar
    check "internal error in p-apostrophe" scan [string index $::Ld 0] %c z
    dcl-type $z int
    lex_getchar
    check "closing apostrophe expected, got $::Ld" eq $::Ld \'
    lex_black
    return $z
}

proc p-number {} {
	set z [expr {0+$::Ld}]
	dcl-type $z int
	lex_black
	return $z
}
proc p-null {} {
	lex_black
	return NULL
}
proc p-subscript {name} { ;# to be used in while loop

    check "operator $::Ld not allowed after vector (* $name *)" expr { $::Lt eq "lp" || $::Lt eq "ls" }

    set i [p-pexpr]
    do-subscript $name $i
}
proc do-subscript {a i} {
    set atype $::Type($a)
    check "subscript not integer" is-intlike $::Type($i)
    set ctype [mangle $atype]
    set etype [element-type $atype]
    set cetype [mangle $etype]

#code "/* ////////// a $atype c $ctype e $etype ce $cetype */"

    set name sub_[serial]
    new-var $name _NOTUSED_ "$cetype* $name= ($cetype*)NULL;"
    dcl-type *$name $etype
    code "  err= ${ctype}_Subscript ( world, &$name, &$a, $i );"
    code "  [if-err-jump  "(* internal subscript expression:   $a \[$i\]   *)" ]"
    return *$name
}
proc p-subfield {name} { ;# to be used in while loop

    if { $::Lt=="lp" } {
	p-get "("
	set f [p-get-ident]
	check "expected close paren after subscript, got $::Ld" eq $::Lt rp ;# must be \)
	p-get ")"
    } elseif { $::Ld eq "." || $::Ld eq "->" } {
	p-get $::Ld
	check "expected identifier for fieldname after dot, got $::Ld" eq $::Lt id
	set f $::Ld
	lex_black
    }

    do-subfield $name $f
}
proc do-subfield {a f} {
    set at $::Type($a)
    set mat [mangle $at]

    check "$f not a field of $at" info exists ::MemType($at.$f)
    set etype $::MemType($at.$f)

    code "if ( ! $a ) { err= p7set_error_static(world, \"dereferencing NULL p7class pointer: ($at)->$f\"); goto final; }"

    set name $a->m_$f
    dcl-type $name $etype
    return $name
}


################
proc emit-init-func {name} {
    set title [string totitle $name]
    if { $::debugmode eq "-g" } {
	    regsub {[0-9]*$} $title _g_Init title
    } else {
	    regsub {[0-9]*$} $title _Init title
    }

    puts "p7err $title ( Tcl_Interp* interp )"
    puts "\{"
    puts "   p7err err= TCL_OK;"
    puts "   Tcl_Command p;"
    puts "   p7world world= (p7world) p7malloc( \"WORLD\", sizeof (struct struct_p7world) );"
    puts "   assert (world);"
    puts "   assert (interp);"
    puts "#ifdef USE_TCL_STUBS"
    puts "   Tcl_InitStubs( interp, \"8.4\", 0 );"
    puts "#endif"

    puts "   world->interp= interp;"
    puts "#ifdef P7_WORLDMAGIC"
    puts "   world->magic= P7_WORLDMAGIC;"
    puts "#endif"
    puts ""
    puts "   world->globals= (p7globals) p7malloc( \"GLOBALS\", sizeof (struct struct_p7globals) );"
    puts "   /* p7malloc does memset 0, but we are paranoid. */"
    puts "   memset( world->globals, 0, sizeof (struct struct_p7globals) );"
    puts ""
    puts "   world->num_literals= 1 + $::NextLiteral;"
    puts "   world->literals= (p7tcl*) p7malloc( \"LITERALS\", world->num_literals * sizeof (p7tcl) );"
    puts "   /* p7malloc does memset 0, but we are paranoid. */"
    puts "   memset( world->literals, 0, world->num_literals * sizeof (p7tcl) );"
    puts ""
    puts "   world->literals\[ $::LIT_EMPTY \]= Tcl_NewStringObj( \"\", -1 );"
    puts "   p7tcl_Link( world->literals\[ $::LIT_EMPTY \] );"
    puts ""
    puts "   world->literals\[ $::LIT_ERRORINFO \]= Tcl_NewStringObj( \"errorInfo\", -1 );"
    puts "   p7tcl_Link( world->literals\[ $::LIT_ERRORINFO \] );"
    puts ""
    puts "   world->literals\[ $::LIT_ERRORCODE \]= Tcl_NewStringObj( \"errorCode\", -1 );"
    puts "   p7tcl_Link( world->literals\[ $::LIT_ERRORCODE \] );"
    puts ""
    puts "   p7world_link(world);"
    puts ""
    puts "   char* tcl_pkg= (char*) Tcl_PkgRequire( interp, \"Tcl\", \"8.4\", 0 );"
    puts "   if (!tcl_pkg) { err=TCL_ERROR; goto final; }"
    puts ""
    puts "   Tcl_PkgProvide( interp, \"$name\", \"1.0\" );"
    puts ""
    puts "   fsync(2);"
#   foreach p $::Pprocs {
#       puts "   p= Tcl_CreateObjCommand( interp, \"$p\", p7wrap_$p, (ClientData)world, p7world_unlink );"
#       puts "   if (!p) { err=TCL_ERROR; goto final; }"
#       puts "   p7world_link(world);"
#   }
    foreach p [array names ::Counts] {
        puts "   p= Tcl_CreateObjCommand( interp, \"$p\", [identifize p7megawrap_$p], (ClientData)world, p7world_unlink );"
        puts "   if (!p) { err=TCL_ERROR; goto final; }"
        puts "   p7world_link(world);"
    }
    foreach p $::Pprocs {
	if {[string match *_INIT_0___void $p]} {
	    puts "  err= p7func_$p ( world );" 
	    puts "   if (err) { goto final; }"
	}
    }
    puts "final:"
    puts "  p7world_unlink( (ClientData)world );"
    puts "  return TCL_OK;"
    puts "\}"
}
    
################
proc master-p7source {filename args} {
    # protection against ".." comes along with "/." and "^."
    if {[regexp {^[.]} $filename]} {error "cannot p7source dot-files: $filename"}
    if {[regexp {/[.]} $filename]} {error "cannot p7source dot-files: $filename"}
    # require (-nocase) .tcl or .p7 extension
    if {![regexp -nocase {[.](tcl|p7i?)$} $filename]} {error "p7source requires .tcl or .p7 or .p7i extension: $filename"}

    set ix [list $filename [lrange $args 0 end]]
    if {![ie ::Sourced($ix)]} {
	set ::Sourced($ix) 1
	set fd [open $filename]
puts stderr "== Sourcing $filename {$args}"
	set guts [read $fd]
	set guts [string map $args $guts]
	Slave eval $guts
	close $fd
    }
}
################

set ::LIT_EMPTY 0
set ::LIT_ERRORINFO 1
set ::LIT_ERRORCODE 2
set ::NextLiteral 2    ;# actually Last not Next

set ::SERIAL 0
set ::NextClsNum 0
set ::ContinueLabel undefined
set ::DefineP7ProcList {}
set ::FinalLabel "final"

switch -exact [llength $argv] {
	1 { 
		set debugmode ""
		set mainfile [lindex $argv 0]
	}
	2 { 
		set debugmode [lindex $argv 0]
		set mainfile [lindex $argv 1]
	}
	default {
		puts stderr "Usage:  tclsh $argv0  [-g]  sourcefile.p7 > ___tmpfile.c"
		exit 13
	}
}

puts "/* generated by $argv0 $debugmode $mainfile */"
puts "#include <parenthetcl.h>"

interp create -safe -- Slave
interp alias Slave p7class	{} master-p7class 
interp alias Slave p7proc	{} master-p7proc 
interp alias Slave p7global	{} master-p7global 
interp alias Slave p7source	{} master-p7source

master-p7source $mainfile

# predefined globals
master-p7global (tcl)p7errorInfo
master-p7global (tcl)p7errorCode

generate-p7globals
emit-p7check

foreach {name paramlist body} $::DefineP7ProcList {
    define-p7proc $name $paramlist $body
}

#################
regsub {[.]\w+$} $mainfile {} load_name
regsub {^.*[/]} $load_name {} load_name
emit-megawrappers
emit-init-func $load_name

emit-dyna-all
exit 0

#END $Id: p7translate.tcl,v 1.2 2006/05/07 04:56:18 strick Exp strick $
